"""A collection of builder functions for models usable with different envs in PPO."""
import inspect
import itertools
from collections import OrderedDict
from functools import partial
from typing import Callable

# registry for base models
MODELS_AVAILABLE = {}
MODEL_BUILDERS = {}


def register_model(model_name):
    def inner(func):
        assert all(p in inspect.signature(func).parameters for p in ["env", "distribution", "bs", "sequence_length"]), \
            (f"Model builder {func.__name__} must have the following parameters: env, distribution, bs, sequence_length"
             f"but has {inspect.signature(func).parameters}")

        # check that no parts of the model name are in the arguments later added
        assert all(p not in model_name for p in ["build", "models", "ffn", "rnn",
                                                 "lstm", "gru", "shared", "distinct", "blind"]), \
            ("Model name must not contain any of the following reserved name parts: "
             "build, models, ffn, rnn, lstm, gru, shared, distinct, blind")

        MODELS_AVAILABLE[model_name] = func

        arguments = OrderedDict(sorted([
            ("model_type", ["ffn", "rnn", "lstm", "gru"]),
            ("blind", [True, False]),
            ("shared", [True, False]),
        ], key=lambda x: x[0]))

        combinations = list(itertools.product(*arguments.values()))
        parameter_dicts = list(map(lambda c: dict(zip(arguments.keys(), c)), combinations))

        for pd in parameter_dicts:
            name_parts = pd.copy()
            name_parts["shared"] = "shared" if name_parts["shared"] else "distinct"
            name_parts["blind"] = "blind" if name_parts["blind"] else ""

            func_name = f"build_{model_name}_{'_'.join([str(val) for n, val in sorted(name_parts.items(), key=lambda x: x[0]) if val])}_models"
            MODEL_BUILDERS[func_name] = partial(func, **pd)
            MODEL_BUILDERS[func_name].__name__ = func_name

            # backwards compatibility
            name_parts["model"] = model_name
            func_name = f"build_{'_'.join([str(val) for n, val in sorted(name_parts.items(), key=lambda x: x[0]) if val])}_models"
            MODEL_BUILDERS[func_name] = partial(func, **pd)
            MODEL_BUILDERS[func_name].__name__ = func_name

        return func

    return inner


def get_model_shortname(model_builder_name: str) -> str:
    """Get short name of model builder.

    Accounts for underscores in the model name.

    Args:
        model_builder_name: Name of model builder function.

    Returns:
        Short name of model builder function.
    """

    # split name into parts
    parts = model_builder_name.split("_")

    # remove "build" and "models", and all parts related to arguments set in register_model
    parts = [p for p in parts if
             p not in ["build", "models", "ffn", "rnn", "lstm", "gru", "shared", "distinct", "blind"]]

    # merge parts to build short name
    return "_".join(parts)


def get_model_type(model_builder) -> str:
    """Get short name of type of model builder."""
    return model_builder.keywords["model_type"]


def get_model_builder(model="simple", model_type: str = "ffn", shared: bool = False, blind: bool = True) -> Callable:
    """Get a builder function for a model with the described parameters."""
    # TODO shared seems not to work yet
    params = locals()
    params["shared"] = "shared" if params["shared"] else "distinct"
    params["blind"] = "blind" if params["blind"] else ""
    params.pop("model")

    params = [str(val) for key, val in sorted(params.items(), key=lambda x: x[0]) if val != ""]
    return MODEL_BUILDERS[f"build_{model}_{'_'.join(params)}_models"]



import json
import os

from matplotlib import pyplot as plt

from angorapy.common.const import PATH_TO_EXPERIMENTS, QUALITATIVE_COLOR_PALETTE

experiment_ids = ['1653053413', '1655284851', '1654708464']

reward_developments = {}
for id in experiment_ids:
    with open(os.path.join("../../../", PATH_TO_EXPERIMENTS, str(id), "meta.json")) as f:
        meta = json.load(f)
    with open(os.path.join("../../../", PATH_TO_EXPERIMENTS, str(id), "progress.json")) as f:
        progress = json.load(f)

    exp_name = meta["hyperparameters"]["distribution"]
    reward_developments[exp_name] = progress["rewards"]["mean"]

for i, (name, rewards) in enumerate(reward_developments.items()):
    plt.plot(rewards[:800], label=name, color=QUALITATIVE_COLOR_PALETTE[i])

plt.title("In-Hand Object Manipulation")
plt.xlabel("Cycle")
plt.ylabel("Avg. Episode Return")
plt.legend()

plt.xlim(0, 800)
plt.ylim(0)

plt.gcf().set_size_inches(16, 4)
plt.show()
# plt.savefig("../../../docs/figures/manipulate-progress.pdf", format="pdf", bbox_inches="tight")


#!/usr/bin/env python
"""Evaluate a loaded agent on a task."""

import argparse
import json
import os
import sys
import statsmodels.stats.api as sms
from tqdm import tqdm

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'

from angorapy.tasks.hand.shadowhand import BaseShadowHandEnv
from angorapy.utilities.datatypes import mpi_condense_stats
import statistics
import time

from angorapy.agent.ppo_agent import PPOAgent
from angorapy.common.const import BASE_SAVE_PATH

try:
    from mpi4py import MPI
except ImportError:
    MPI = None

import tensorflow as tf

physical_devices = tf.config.list_physical_devices('GPU')
tf.config.experimental.set_memory_growth(physical_devices[0], enable=True)

tf.get_logger().setLevel('INFO')

parser = argparse.ArgumentParser(description="Evaluate an agent.")
parser.add_argument("id", type=int, nargs="?", help="id of the agent, defaults to newest", default="1673350499432390")
parser.add_argument("-n", type=int, help="number of evaluation episodes", default=12)
parser.add_argument("--act-confidently", action="store_true", help="act deterministically without stochasticity")
args = parser.parse_args()

if MPI is None:
    mpi_comm = MPI.COMM_WORLD
    rank = mpi_comm.rank
    is_root = rank == 0
    comm_size = mpi_comm.size
else:
    is_root = True
    rank = 0
    comm_size = 1

if args.id is None:
    ids = map(int, os.listdir(BASE_SAVE_PATH))
    args.id = max(ids)

start = time.time()
agent = PPOAgent.from_agent_state(args.id, "best", path_modifier="../")

# if isinstance(agent.env.unwrapped, BaseShadowHandEnv):
#     agent.env.env.set_delta_t_simulation(0.0005)

if is_root:
    print(f"Agent {args.id} successfully loaded from state 'best' (training performance: {agent.cycle_reward_history[-1]}).")


# determine number of repetitions on this worker
worker_base, worker_extra = divmod(args.n, comm_size)
worker_split = [worker_base + (r < worker_extra) for r in range(comm_size)]
workers_n = worker_split[rank]

stat_bundles, _ = agent.evaluate(workers_n, act_confidently=args.act_confidently)
stats = mpi_condense_stats([stat_bundles])

if is_root:
    average_reward = round(statistics.mean(stats.episode_rewards), 2)
    average_length = round(statistics.mean(stats.episode_lengths), 2)
    std_reward = round(statistics.stdev(stats.episode_rewards), 2)
    std_length = round(statistics.stdev(stats.episode_lengths), 2)

    print(f"Evaluated agent on {stats.numb_completed_episodes} x {agent.env_name} and achieved an average reward of {average_reward} [std: {std_reward}; "
          f"between ({min(stats.episode_rewards)}, {max(stats.episode_rewards)})].\n ")

    print("Auxiliary Performance Measures:")
    for aux_perf_name, aux_perf_trace in stats.auxiliary_performances.items():
        aux_perf_stats = sms.DescrStatsW(aux_perf_trace)
        average_perf = round(aux_perf_stats.mean, 2)
        confidence_interval = round(aux_perf_stats.tconfint_mean()[0], 2), round(aux_perf_stats.tconfint_mean()[1], 2)
        print(f"\t{aux_perf_name}: {average_perf} [{confidence_interval}]")
    print("\n")

    print(f"The agent {'acted confidently' if args.act_confidently else 'acted stochastically.'}\n"
          f"An episode on average took {average_length} steps [std: {std_length}; "
          f"between ({min(stats.episode_lengths)}, {max(stats.episode_lengths)})].\n"
          f"This took me {round(time.time() - start, 2)}s.")

    with open(f"../{agent.experiment_directory}/evaluation.json", "w") as f:
        json.dump(stats._asdict(), f)



import json
import re

import matplotlib
import numpy as np
from matplotlib import pyplot as plt
from angorapy.common.const import QUALITATIVE_COLOR_PALETTE as COLORS
import statsmodels.stats.api as sms

from angorapy.utilities.util import stack_dicts

font = {'family' : 'Times',
        'weight' : 'normal',
        'size'   : 12}

matplotlib.rc('font', **font)

partial_results = []
for filename in ["./storage/predictability_repeated_august19d.json", "./storage/predictability_repeated_august19e.json"]:
    with open(filename, "r") as f:
        partial_results.append(json.load(f))

results = {}
for key in partial_results[0].keys():
    for target in partial_results[0][key].keys():
        for part in partial_results:
            if key not in part or target not in part[key]:
                continue

            if key not in results:
                results[key] = {}
            if target not in results[key]:
                results[key][target] = []

            results[key][target] = results[key][target] + part[key][target]

# remove entries from results that are contain no underscore in their name
results = {k: v for k, v in results.items() if "_" in k or "goal" in k}

x = results.keys()

# create x labels by removing 'activation' potentially surrounded by underscores from the x label name
x_labels = [re.sub(r"(_)?internal(_)?", "", s) for s in results.keys()]

# rename x label "pmc_recurrent_layer" to "PMC"
x_labels = [re.sub(r"pmc_recurrent_layer", "PMC", s) for s in x_labels]
x_labels = [s.upper() if s != "goal" else "Goal" for s in x_labels]

targets = ["reward", "goals_achieved_so_far"]
target_names = ["Reward", "Goals Achieved So Far"]

# create a figure with a single plot
fig, axs = plt.subplots(1, len(targets), sharey="row", figsize=(8, 2))

# plot barplots for all targets, with bars placed next to each other not on top of each other
for i, target in enumerate(targets):
    # plot barplot for target
    axs[i].bar(np.arange(len(x)),
            [np.mean(results[s][target]) for s in results.keys()],
            # use confidence interval as error bars
            yerr=[sms.DescrStatsW(results[s][target]).tconfint_mean()[1] - np.mean(results[s][target]) for s in results.keys()],
            capsize=2,
            color=COLORS[i])

    # set labels of the x axis to keys in results
    axs[i].set_xticks(np.arange(len(x)))
    axs[i].set_xticklabels(x_labels)

    # make x axis labels vertical
    axs[i].tick_params(axis='x', rotation=90)

    if i == 0:
        axs[i].set_ylabel("Predictability (R2)")
    else:
        axs[i].tick_params(axis='y', which='both', left=False, labelleft=False)

    axs[i].set_ylim(0, 1)
    axs[i].set_title(target_names[i])

# remove space between subplots
plt.subplots_adjust(wspace=0.05)

plt.savefig(f"predictability_reward_goals.pdf", format="pdf", bbox_inches="tight")


import glob
import os
import json

experiment_dir = "../../../storage/experiments/"

experiment_group_inclusion_filters = {
    # all filters are "only if in" filters, i.e. if the experiment has any of the values in the filter, it will be
    # included, otherwise it will be excluded

    "group_one": {
        "architecture": ["build_fpn_v2_models"],
    },

    "group_two": {
        "architecture": ["build_fpn_v3_models"],
    },
}

# get ids of all experiments in the storage by globbing the experiment folders
experiment_ids = sorted([os.path.basename(path) for path in glob.glob(f"{experiment_dir}*")
                  if os.path.isfile(path + "/meta.json")])

# load experiment group data
group_ids = {}
for group_name, group_filters in experiment_group_inclusion_filters.items():
    group_ids[group_name] = []
    for experiment_id in experiment_ids:
        try:
            with open(f"{experiment_dir}{experiment_id}/meta.json", "r") as f:
                experiment = json.load(f)
        except json.JSONDecodeError as e:
            # print(f"Skipping experiment {experiment_id} because it has corrupted meta.json!")
            continue

        include_experiment = True
        for filter_name, filter_values in group_filters.items():
            if experiment["hyperparameters"][filter_name] not in filter_values:
                # print(f"Excluding experiment {experiment_id} because it does not match filter {filter_name}: {experiment['hyperparameters'][filter_name]}!={filter_values}")
                include_experiment = False
                break

        if include_experiment:
            group_ids[group_name].append(experiment_id)

# plot it



import json
import os
import re
from json import JSONDecodeError

import matplotlib
import matplotlib.pyplot as plt
import numpy as np

from angorapy.common.const import PATH_TO_EXPERIMENTS, BASE_SAVE_PATH

matplotlib.use('TkAgg')

group_names = ["benchmark-performance-pdl", "benchmark-performance-llc", "benchmark-performance-cp",
               "benchmark-performance-ab"]
titles = ["Pendulum", "LunarLanderContinuous", "CartPole", "Acrobot"]

# group_names = [
#     "benchmark-performance-ant",
#     "benchmark-performance-walker2d",
#     "benchmark-performance-swimmer",
#     "benchmark-performance-reacher",
#     "benchmark-performance-humanoidstandup",
#     "benchmark-performance-humanoid",
#     "benchmark-performance-hopper",
#     "benchmark-performance-halfcheetah"
# ]

# group_names = [
#     "benchmark-beta-reach",
#     "benchmark-beta-freereach"
# ]

# titles = [n.split("-")[-1].capitalize() for n in group_names]

exp_dir = "../../../" + PATH_TO_EXPERIMENTS
experiment_paths = [os.path.join(exp_dir, p) for p in os.listdir(exp_dir)]

environments = {}
reward_thresholds = {}
experiments_by_groups = {}
envs_available = set()

for exp_path in experiment_paths:

    eid_m = re.match("[0-9]+", str(exp_path.split("/")[-1]))
    if eid_m:
        eid = eid_m.group(0)
        model_path = os.path.join(BASE_SAVE_PATH, eid)

        if os.path.isfile(os.path.join(exp_path, "progress.json")):
            with open(os.path.join(exp_path, "progress.json"), "r") as f:
                progress = json.load(f)

            with open(os.path.join(exp_path, "meta.json"), "r") as f:
                try:
                    meta = json.load(f)
                except JSONDecodeError as jserr:
                    continue

            exp_group = meta.get("experiment_group", "n/a")

            if exp_group not in group_names:
                continue

            reward_threshold = None if meta["environment"]["reward_threshold"] == "None" else float(
                meta["environment"]["reward_threshold"])

            if not exp_group in experiments_by_groups.keys():
                experiments_by_groups[exp_group] = {}
                reward_thresholds[exp_group] = reward_threshold
                environments[exp_group] = meta["environment"]["name"]

            envs_available.add(meta["environment"]["name"])

            experiments_by_groups[exp_group].update({
                eid: progress
            })

n_rows, n_cols = 1, 4
fig, axs = plt.subplots(n_rows, n_cols)
fig.set_size_inches(16, 3 * n_rows)

if not isinstance(axs[0], list):
    axs = [axs]

for i, name in enumerate(group_names):
    data = experiments_by_groups[name]
    reward_trajectories = list(map(lambda x: x["rewards"]["mean"], data.values()))
    max_length = max([len(x) for x in reward_trajectories])
    padded_reward_trajectories = list(map(lambda x: np.pad(x, (0, max_length - len(x)),
                                                           mode="constant",
                                                           constant_values=np.nan), reward_trajectories))
    mean_reward = np.ma.mean(np.ma.array(padded_reward_trajectories, mask=np.isnan(padded_reward_trajectories)), axis=0)
    std_reward = np.ma.std(np.ma.array(padded_reward_trajectories, mask=np.isnan(padded_reward_trajectories)), axis=0)

    ax = axs[i // n_cols][i % n_cols]

    ax.plot(mean_reward)
    ax.fill_between(range(mean_reward.shape[0]), mean_reward - std_reward, mean_reward + std_reward, alpha=.2)

    ax.set_xlim(0, mean_reward.shape[0] - 1)
    ax.set_ylim(np.min(mean_reward - std_reward), np.max(mean_reward + std_reward))
    ax.set_xlabel("Cycle")
    ax.set_title(titles[i])

    if i % n_cols == 0:
        ax.set_ylabel("Episode Return")

plt.savefig(f"../../../docs/figures/benchmarks/{'_'.join(titles)}_benchmark.pdf", format="pdf", bbox_inches='tight')
plt.show()



from angorapy.analysis.investigators.base_investigator import Investigator
from angorapy.analysis.investigators.predictability import Predictability
# from angorapy.analysis.investigators.dynamics import Dynamics, LatentDynamics
from angorapy.analysis.investigators.correlation import Correlation
from angorapy.analysis.investigators.activity_autocorrelation import ActivityAutocorrelation


from typing import Union

import gymnasium as gym

from angorapy.common.postprocessors import StateNormalizer, RewardNormalizer, BasePostProcessor
from angorapy.tasks.wrappers import TaskWrapper, TransformationWrapper
from angorapy.utilities.core import env_extract_dims


def make_task(env_name,
              reward_config: Union[str, dict] = None,
              reward_function: Union[str, dict] = None,
              transformers=None,
              **kwargs):
    """Make environment, including a possible reward config and transformers.

    If transformers is None, the default transformers are used, i.e. StateNormalizationTransformer and
    RewardNormalizationTransformer.

    Args:
        env_name (str): The registered name of the environment to make.
        reward_config (Union[str, dict]): The reward config to use.
        reward_function (Union[str, dict]): The reward function to use.
        transformers (Optional[List[BaseTransformer]]): The transformers to use.

    Returns:
        TaskWrapper: A wrapped instance of the task.
    """
    if transformers is None:
        transformers = [StateNormalizer, RewardNormalizer]

    base_env = gym.make(env_name, **kwargs)
    state_dim, n_actions = env_extract_dims(base_env)

    if transformers is None:
        transformers = []
    elif all(isinstance(t, BasePostProcessor) for t in transformers):
        transformers = transformers
    elif all(callable(t) for t in transformers):
        transformers = [t(env_name, state_dim, n_actions) for t in transformers]

    if len(transformers) > 0:
        env = TransformationWrapper(base_env, transformers=transformers)
    else:
        env = TaskWrapper(base_env)

    if reward_function is not None and hasattr(env, "reward_function"):
        env.set_reward_function(reward_function)

    if reward_config is not None and hasattr(env, "reward_config"):
        env.set_reward_config(reward_config)

    return env


make_env = make_task  # alias for backwards compatibility



import os
import re
import sys

import gymnasium as gym

os.environ['MPLCONFIGDIR'] = os.getcwd() + "/configs/"

import pprint
import traceback

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from angorapy.utilities.defaults import autoselect_distribution

import distance
import numpy as np

import tensorflow as tf

import argparse
import logging

import argcomplete
from gymnasium.spaces import Box

from angorapy.configs import hp_config
from angorapy.common.policies import get_distribution_by_short_name
from angorapy.models import get_model_builder, MODELS_AVAILABLE
from angorapy.common.const import COLORS
from angorapy.utilities.monitoring import Monitor
from angorapy.utilities.core import env_extract_dims, mpi_print
from angorapy import make_task
from angorapy.common.postprocessors import StateNormalizer, RewardNormalizer
from angorapy.agent.ppo_agent import PPOAgent

try:
    from mpi4py import MPI
except ImportError:
    MPI = None


class InconsistentArgumentError(Exception):
    """Arguments given to a process were inconsistent."""
    pass


def run_experiment(environment, settings: dict, verbose=True, use_monitor=False):
    """Run an experiment with the given settings ."""
    if settings["cpu"]:
        print("Deactivating GPU")
        tf.config.set_visible_devices([], "GPU")
        os.environ["CUDA_VISIBLE_DEVICES"] = "-1"

    # setup MPI
    if MPI is not None:
        mpi_rank = MPI.COMM_WORLD.rank
        is_root = mpi_rank == 0
    else:
        mpi_rank = 0
        is_root = True

    # sanity checks and warnings for given parameters
    if settings["preload"] is not None and settings["load_from"] is not None:
        raise InconsistentArgumentError("You gave both a loading from a pretrained component and from another "
                                        "agent state. This cannot be resolved.")

    # determine relevant transformers
    wrappers = []
    wrappers.append(StateNormalizer) if not settings["no_state_norming"] else None
    wrappers.append(RewardNormalizer) if not settings["no_reward_norming"] else None

    # setup environment and extract and report information
    env = make_task(environment,
                    reward_config=settings["rcon"],
                    transformers=wrappers,
                    render_mode="rgb_array" if re.match(".*[Vv]is(ion|ual).*", environment) else None)
    state_dim, number_of_actions = env_extract_dims(env)

    if env.spec.max_episode_steps is not None and env.spec.max_episode_steps > settings["horizon"] and not settings[
        "eval"] and is_root:
        logging.warning("Careful! Your horizon is lower than the max environment steps, "
                        "this will most likely skew stats heavily.")

    # choose and make policy distribution
    if settings["distribution"] is None:
        distribution = autoselect_distribution(env)
        settings["distribution"] = distribution.short_name
    else:
        distribution = get_distribution_by_short_name(settings["distribution"])(env)

    # setting appropriate model building function
    try:
        blind = not ("vision" in list(env.observation_space["observation"].keys())
                     and len(env.observation_space["observation"]["vision"].shape) > 1)
    except:
        blind = True
    build_models = get_model_builder(model=settings["architecture"], model_type=settings["model"],
                                     shared=settings["shared"],
                                     blind=blind)

    # announce experiment
    bc, ec, wn = COLORS["HEADER"], COLORS["ENDC"], COLORS["WARNING"]
    if verbose and is_root:
        env_action_space_type = "continuous" if isinstance(env.action_space, Box) else "discrete"
        env_observation_space_type = "continuous" if isinstance(env.observation_space, Box) else "discrete"
        env_name = env.unwrapped.spec.id

        print(f"-----------------------------------------\n"
              f"{wn}Learning the Task{ec}: {bc}{env_name}{ec}\n"
              f"{bc}{state_dim}{ec}-dimensional states ({bc}{env_observation_space_type}{ec}) "
              f"and {bc}{number_of_actions}{ec} actions ({bc}{env_action_space_type}{ec}).\n"
              f"Config: {settings['pcon']}\n"
              f"Model: {build_models.__name__}\n"
              f"Distribution: {settings['distribution']}\n"
              f"-----------------------------------------\n")

        print(f"{wn}HyperParameters{ec}: {pprint.pformat(settings)}\n")

    if settings["load_from"] is not None:
        if verbose and is_root:
            print(f"{wn}Loading{ec} from state {settings['load_from']}")
        agent = PPOAgent.from_agent_state(settings["load_from"], from_iteration="last",
                                          n_optimizers=settings["n_optimizers"])
    else:
        # set up the agent and a reporting module
        agent = PPOAgent(build_models, env, horizon=settings["horizon"], workers=settings["workers"],
                         learning_rate=settings["lr_pi"], discount=settings["discount"], lam=settings["lam"],
                         clip=settings["clip"], c_entropy=settings["c_entropy"], c_value=settings["c_value"],
                         gradient_clipping=settings["grad_norm"], clip_values=settings["clip_values"],
                         tbptt_length=settings["tbptt"], lr_schedule=settings["lr_schedule"], distribution=distribution,
                         reward_configuration=settings["rcon"], debug=settings["debug"],
                         pretrained_components=None if settings["preload"] is None else [settings["preload"]],
                         n_optimizers=settings["n_optimizers"])

        if is_root:
            print(f"{wn}Created agent{ec} with ID {bc}{agent.agent_id}{ec}")

        # load pretrained components
        if settings["component_dir"] is not None:
            agent.load_components(settings["component_dir"])

        # freeze components
        if settings["freeze_components"] is not None:
            for component in settings["freeze_components"]:
                agent.freeze_component(component)

    if len(tf.config.list_physical_devices('GPU')) > 0:
        agent.set_gpu(not settings["cpu"])
    else:
        print("No GPU found, running on CPU.")
        agent.set_gpu(False)

    monitor = None
    if use_monitor and is_root:
        monitor = Monitor(agent,
                          agent.env,
                          frequency=settings["monitor_frequency"],
                          gif_every=settings["gif_every"],
                          id=agent.agent_id,
                          iterations=settings["iterations"],
                          config_name=settings["pcon"],
                          experiment_group=settings["experiment_group"])

    try:
        # tf.profiler.experimental.start("logdir")
        agent.drill(n=settings["iterations"], epochs=settings["epochs"], batch_size=settings["batch_size"],
                    monitor=monitor, save_every=settings["save_every"], separate_eval=settings["eval"],
                    stop_early=settings["stop_early"], radical_evaluation=settings["radical_evaluation"])
        # tf.profiler.experimental.stop()
    except KeyboardInterrupt:
        mpi_print("Ended by user.")
    except Exception:
        if mpi_rank == 0:
            traceback.print_exc()
    finally:
        if mpi_rank == 0:
            agent.finalize()

    agent.save_agent_state()
    env.close()

    return agent


if __name__ == "__main__":
    tf.get_logger().setLevel('INFO')
    all_envs = [e.id for e in list(gym.envs.registry.values())]

    # parse commandline arguments
    parser = argparse.ArgumentParser(description="Train a PPO Agent on some task.")

    # general parameters
    parser.add_argument("env", nargs='?', type=str, default="ReachAbsolute-v0", help="the target gym environment")
    parser.add_argument("--architecture", choices=MODELS_AVAILABLE, default="simple",
                        help="architecture of the policy")
    parser.add_argument("--model", choices=["ffn", "rnn", "lstm", "gru"], default="ffn",
                        help=f"model type if architecture allows for choices")
    parser.add_argument("--distribution", type=str, default=None,
                        choices=["categorical", "gaussian", "beta", "rbeta", "multi-categorical"])
    parser.add_argument("--shared", action="store_true",
                        help=f"make the model share part of the network for policy and value")
    parser.add_argument("--iterations", type=int, default=5000, help=f"number of iterations before training ends")

    # meta arguments
    parser.add_argument("--pcon", type=str, default=None, help="config name (utilities/hp_config.py) to be loaded")
    parser.add_argument("--rcon", type=str, default=None,
                        help="config (utilities/reward_config.py) of the reward function")
    parser.add_argument("--experiment-group", type=str, default="default", help="experiment group identifier")
    parser.add_argument("--cpu", action="store_true", help=f"use cpu only")
    parser.add_argument("--sequential", action="store_true", help=f"run worker sequentially workers")
    parser.add_argument("--gi-from", type=int, default=None, help=f"load from given agent id")
    parser.add_argument("--preload", type=str, default=None, help=f"load visual component weights from pretraining")
    parser.add_argument("--component-dir", type=str, default=None, help=f"path to pretrained components")
    parser.add_argument("--export-file", type=int, default=None, help=f"save policy to be loaded in workers into file")
    parser.add_argument("--eval", action="store_true", help=f"evaluate additionally to have at least 5 eps")
    parser.add_argument("--radical-evaluation", action="store_true", help=f"only record stats from seperate evaluation")
    parser.add_argument("--save-every", type=int, default=0, help=f"save agent every given number of iterations")
    parser.add_argument("--monitor-frequency", type=int, default=1, help=f"update the monitor every n iterations.")
    parser.add_argument("--gif-every", type=int, default=0, help=f"make a gif every n iterations.")
    parser.add_argument("--debug", action="store_true", help=f"run in debug mode (eager mode)")
    parser.add_argument("--no-monitor", action="store_true", help="dont use a monitor")
    parser.add_argument("--freeze-components", type=str, nargs="+", help=f"Components to freeze the weights of.")
    parser.add_argument("--n-optimizers", type=int, default=None, help=f"number of optimizers; "
                                                                       f"default is all GPUs or CPUs (if no GPUs found)")

    # gathering parameters
    parser.add_argument("--workers", type=int, default=8, help=f"the number of workers exploring the environment")
    parser.add_argument("--horizon", type=int, default=2048,
                        help=f"number of time steps one worker generates per cycle")
    parser.add_argument("--discount", type=float, default=0.99, help=f"discount factor for future rewards")
    parser.add_argument("--lam", type=float, default=0.97, help=f"lambda parameter in the GAE algorithm")
    parser.add_argument("--no-state-norming", action="store_true", help=f"do not normalize states")
    parser.add_argument("--no-reward-norming", action="store_true", help=f"do not normalize rewards")

    # optimization parameters
    parser.add_argument("--epochs", type=int, default=3, help=f"the number of optimization epochs in each cycle")
    parser.add_argument("--batch-size", type=int, default=64, help=f"minibatch size during optimization")
    parser.add_argument("--lr-pi", type=float, default=1e-3, help=f"learning rate of the policy")
    parser.add_argument("--lr-schedule", type=str, default=None, choices=[None, "exponential"],
                        help=f"lr schedule type")
    parser.add_argument("--clip", type=float, default=0.2, help=f"clipping range around 1 for the objective function")
    parser.add_argument("--c-entropy", type=float, default=0.01, help=f"entropy factor in objective function")
    parser.add_argument("--c-value", type=float, default=1, help=f"value factor in objective function")
    parser.add_argument("--tbptt", type=int, default=16, help=f"length of subsequences in truncated BPTT")
    parser.add_argument("--grad-norm", type=float, default=0.5, help=f"norm for gradient clipping, 0 deactivates")
    parser.add_argument("--clip-values", action="store_true", help=f"clip value objective")
    parser.add_argument("--stop-early", action="store_true", help=f"stop early if threshold of env was surpassed")

    if MPI is not None:
        comm = MPI.COMM_WORLD
        rank = comm.Get_rank()
        is_root = rank == 0
    else:
        rank = 0
        is_root = True

    # read arguments
    argcomplete.autocomplete(parser)
    args = parser.parse_args()

    if args.env not in all_envs:
        if is_root:
            indices = np.argsort([distance.levenshtein(w, args.env) for w in all_envs])[:3]
            print(f"Unknown environment {args.env}. Did you mean one of {[all_envs[i] for i in indices]}")
        exit()

    # if config is given load it as default, then overwrite with any goal given parameters
    if args.pcon is not None:
        try:
            parser.set_defaults(**getattr(hp_config, args.pcon))
            args = parser.parse_args()
            if is_root:
                print(f"Loaded Config {args.pcon}.")
        except AttributeError as err:
            raise ImportError("Cannot find config under given name. Does it exist in utilities/hp_config.py?")

    if args.debug:
        tf.config.run_functions_eagerly(True)
        if is_root:
            logging.warning("YOU ARE RUNNING IN DEBUG MODE!")
    else:
        tf.config.run_functions_eagerly(False)

    try:
        run_experiment(args.env, vars(args), use_monitor=not args.no_monitor)
    except Exception as e:
        if rank == 0:
            traceback.print_exc()



"""Probability distributions used in Stochastic Policies."""

import abc
import math
from typing import Union, Tuple, List

import gymnasium as gym
import numpy as np
import scipy
import tensorflow as tf

from angorapy.agent.utils import extract_discrete_action_probabilities
from angorapy.common.const import EPSILON
from angorapy.common.layers import StdevLayer, BetaDistributionSpreadLayer
from angorapy.utilities.core import env_extract_dims


class BasePolicyDistribution(abc.ABC):
    """Abstract base class for policy distributions.

    Each policy implementation provides basic probabilistic methods to calculate probability density and entropy, both
    in standard and in log space. It furthermore gives the characteristic act method that given the probability
    parameters samples the step_tuple and provides the actions probability."""

    def __init__(self, env: gym.Env):
        self.state_dim, self.action_dim = env_extract_dims(env)
        self.action_space: gym.Space = env.action_space

    @property
    def is_continuous(self):
        """Indicate whether the distribution is continuous."""
        return False

    @property
    @abc.abstractmethod
    def short_name(self):
        """Distribution'serialization short name."""
        return "base"

    @property
    @abc.abstractmethod
    def has_log_params(self):
        """Indicate whether the parameters of the distribution are expected to be given in log space."""
        pass

    @abc.abstractmethod
    def act(self, *args, **kwargs):
        """Sample an step_tuple from the distribution and return it alongside its probability."""
        pass

    def act_numpy(self, *args, **kwargs):
        """Can be overwritten by subclasses to provide a faster numpy implementation of the act method. This pure numpy
        implementation is called by the gatherer where no GPU utilization is possible."""
        return self.act(*args, **kwargs)

    @abc.abstractmethod
    def act_deterministic(self, *args, **kwargs) -> Tuple[np.ndarray, np.ndarray]:
        pass

    @abc.abstractmethod
    def sample(self, *args, **kwargs):
        """Sample an step_tuple from the distribution."""
        pass

    @abc.abstractmethod
    def probability(self, *args, **kwargs):
        """Get the probability of given sample with given distribution parameters. For continuous distributions this is
        the probability density function."""
        pass

    @abc.abstractmethod
    def log_probability(self, *args, **kwargs):
        """Get the logarithmic probability of given sample with given distribution parameters. For continuous
        distributions this is the probability density function.

        The logarithmic probability has better numerical stability properties during later calculations."""
        pass

    @abc.abstractmethod
    def entropy(self):
        """Calculate the entropy of the distribution based on log parameters."""
        pass

    @tf.function
    def approximate_kl_divergence(self, log_pa: tf.Tensor, log_pb: tf.Tensor):
        """Approximate KL-divergence between distributions a and b where some sample has probability pa in a
        and pb in b, based on log probabilities."""
        return .5 * tf.reduce_mean(tf.square(log_pa - log_pb))

    @abc.abstractmethod
    def build_action_head(self, n_actions: Tuple[int], input_shape: tuple, batch_size: Union[int, None]):
        """Build the step_tuple head of a policy using this distribution."""
        pass


class CategoricalPolicyDistribution(BasePolicyDistribution):
    """Policy implementation of categorical (also discrete) distributions. That is, this policy is to be used in any
    case where the step_tuple space is discrete and the agent thus predicts a pmf over the possible actions."""

    def act_deterministic(self, log_probabilities: Union[tf.Tensor, np.ndarray]) -> Tuple[np.ndarray, np.ndarray]:
        return tf.squeeze(tf.argmax(log_probabilities, axis=-1)).numpy(), tf.reduce_max(log_probabilities, axis=-1).numpy()

    @property
    def short_name(self):
        """Policy's short identifier."""
        return "categorical"

    @property
    def has_log_params(self):
        """Categorical distribution expects pmf in log space."""
        return True

    def act(self, log_probabilities: Union[tf.Tensor, np.ndarray]) -> Tuple[np.ndarray, np.ndarray]:
        """Sample a step_tuple from a discrete step_tuple distribution predicted by the given policy for a given state.
        """
        action = self.sample(log_probabilities)
        return action, tf.squeeze(log_probabilities)[action]

    def sample(self, log_probabilities):
        """Sample an action from the distribution."""
        assert isinstance(log_probabilities, tf.Tensor) or isinstance(log_probabilities, np.ndarray), \
            f"Policy methods (act_discrete) require Tensors or Numpy Arrays as input, " \
            f"not {type(log_probabilities).__name__}."

        if tf.rank(log_probabilities) == 3:
            # there appears to be a sequence dimension
            assert log_probabilities.shape[
                       1] == 1, "Policy actions can only be selected for a single timestep, but the " \
                                "dimensionality of the given tensor is more than 1 at rank 1."

            log_probabilities = tf.squeeze(log_probabilities, axis=1)

        action = tf.random.categorical(log_probabilities, 1)[0][0]
        return action.numpy()

    def probability(self, **kwargs):
        """Not Implemented"""
        raise NotImplementedError("A categorical distribution has no pdf.")

    @staticmethod
    @tf.function
    def log_probability(actions: tf.Tensor, distribution: tf.Tensor):
        """Calculate log probability of a (batch of) actions given their distribution."""
        return extract_discrete_action_probabilities(distribution, actions)

    @tf.function
    def _entropy_from_pmf(self, pmf: tf.Tensor):
        """Calculate entropy of a categorical distribution from raw pmf."""
        return - tf.reduce_sum(tf.math.log(pmf) * pmf, axis=-1)

    @tf.function
    def _entropy_from_log_pmf(self, pmf: tf.Tensor):
        """Calculate entropy of a categorical distribution, where the pmf is given as log probabilities."""
        return - tf.reduce_sum(tf.exp(pmf) * pmf, axis=-1)

    @tf.function
    def entropy(self, pmf: tf.Tensor):
        """Calculate entropy of a categorical distribution, where the pmf is given as log probabilities."""
        return self._entropy_from_log_pmf(pmf)

    def build_action_head(self, n_actions: Tuple[int], input_shape: tuple, batch_size: int):
        """Build a discrete step_tuple head as a log softmax output layer."""
        assert len(n_actions) == 1, "Categorical distribution expects only a single action dimension. " \
                                    "For other multiple action dimensions use MultiCategoricalPolicyDistribution."

        inputs = tf.keras.Input(batch_shape=(batch_size,) + tuple(input_shape))

        x = tf.keras.layers.Dense(n_actions[0],
                                  kernel_initializer=tf.keras.initializers.Orthogonal(0.01),
                                  bias_initializer=tf.keras.initializers.Constant(0.0))(inputs)
        x = tf.nn.log_softmax(x, name="log_likelihoods")

        return tf.keras.Model(inputs=inputs, outputs=x, name="discrete_action_head")


class MultiCategoricalPolicyDistribution(BasePolicyDistribution):
    """Policy distribution for multi-discrete action spaces."""

    def act_deterministic(self, log_probabilities: Union[tf.Tensor, np.ndarray]) -> Tuple[np.ndarray, np.ndarray]:
        action = tf.argmax(log_probabilities, axis=-1)

        return tf.squeeze(action).numpy(), tf.reduce_max(log_probabilities, axis=-1).numpy()

    @property
    def short_name(self):
        """Policy's short identifier."""
        return "multi-categorical"

    @property
    def has_log_params(self):
        """Categorical distribution expects pmf in log space."""
        return True

    def act(self, log_probabilities: Union[tf.Tensor, np.ndarray]) -> Tuple[np.ndarray, np.ndarray]:
        """Sample an action from the distribution based on the provided log probabilities."""
        action = self.sample(log_probabilities)
        sample_probability = tf.gather_nd(tf.squeeze(log_probabilities),
                                          tf.stack([tf.range(len(action)), action], axis=-1))

        return action.astype(int), tf.math.reduce_sum(sample_probability, axis=-1)

    def sample(self, log_probabilities: tf.Tensor) -> np.ndarray:
        """Sample an action from the distribution."""
        assert isinstance(log_probabilities, tf.Tensor) or isinstance(log_probabilities, np.ndarray), \
            f"Policy methods (act_discrete) require Tensors or Numpy Arrays as input, " \
            f"not {type(log_probabilities).__name__}."

        if tf.rank(log_probabilities) == 4:
            # there appears to be a sequence dimension
            assert log_probabilities.shape[
                       1] == 1, "Policy actions can only be selected for a single timestep, but the " \
                                "dimensionality of the given tensor is more than 1 at rank 1."

            log_probabilities = tf.squeeze(log_probabilities, axis=1)

        action = [tf.random.categorical(log_probabilities[..., a_id, :], 1)[0][0].numpy().item()
                  for a_id in range(log_probabilities.shape[-2])]

        return np.array(action)

    def probability(self, **kwargs):
        """Not Implemented"""
        raise NotImplementedError("A categorical distribution has no pdf.")

    @staticmethod
    @tf.function
    def log_probability(actions, distribution):
        # the tf.where() trick gives us indices of the same shape as the actions tensor
        prepending_indices = tf.cast(tf.where(tf.ones(actions.shape)), dtype=tf.int32)
        indices = tf.concat([prepending_indices, tf.reshape(actions, [-1, 1])], axis=-1)

        single_action_logits = tf.reshape(tf.gather_nd(distribution, indices), actions.shape)

        # reduce by sum to give joined probability of multi-categorical distribution; sum because of log space
        return tf.math.reduce_sum(single_action_logits, axis=-1)

    @tf.function
    def _entropy_from_pmf(self, pmf: tf.Tensor):
        """Calculate entropy of a categorical distribution from raw pmf."""
        return - tf.reduce_sum(tf.math.log(pmf) * pmf, axis=-1)

    @tf.function
    def _entropy_from_log_pmf(self, log_pmf: tf.Tensor):
        """Calculate entropy of a categorical distribution, where the pmf is given as log probabilities."""
        return - tf.reduce_sum(tf.exp(log_pmf) * log_pmf, axis=-1)

    @tf.function
    def entropy(self, pmf: tf.Tensor):
        """Calculate entropy of a categorical distribution, where the pmf is given as log probabilities."""
        return self._entropy_from_log_pmf(pmf)

    def build_action_head(self, n_actions: Tuple[int], input_shape: tuple, batch_size: int):
        """Build a discrete step_tuple head as a log softmax output layer."""
        assert len(n_actions) == 2, "MultiCategoricalPolicyDistribution expects two action dimensions."

        inputs = tf.keras.Input(batch_shape=(batch_size,) + tuple(input_shape))

        x = tf.keras.layers.Dense(tf.multiply(*n_actions),
                                  kernel_initializer=tf.keras.initializers.Orthogonal(0.01),
                                  bias_initializer=tf.keras.initializers.Constant(0.0))(inputs)
        unflattened_shape = x.shape[1:-1].concatenate(n_actions)
        x = tf.keras.layers.Reshape(unflattened_shape)(x)
        x = tf.keras.layers.Activation(tf.nn.log_softmax, name="log_likelihoods", dtype="float32")(x)

        return tf.keras.Model(inputs=inputs, outputs=x, name="discrete_action_head")


class BaseContinuousPolicyDistribution(BasePolicyDistribution, abc.ABC):
    """Abstract base class of continuous policy distributions."""

    def __init__(self, env):
        super().__init__(env)

        self.action_max_values = env.action_space.high
        self.action_min_values = env.action_space.low
        self.action_mm_diff = self.action_max_values - self.action_min_values

    @property
    def is_continuous(self):
        """Indicate that the distribution is continuous."""
        return True


class GaussianPolicyDistribution(BaseContinuousPolicyDistribution):
    """Gaussian Probability Distribution."""

    def act_deterministic(self, means, log_stdevs):
        actions = means
        probabilities = self.log_probability(actions, means=means, log_stdevs=log_stdevs)

        return tf.reshape(actions, [-1]).numpy(), tf.squeeze(probabilities).numpy()

    @property
    def short_name(self):
        """Policy's short identidier."""
        return "gaussian"

    @property
    def has_log_params(self):
        """Gaussian Distribution expects standard deviation in log space"""
        return True

    def act(self, means: tf.Tensor, log_stdevs: tf.Tensor) -> Tuple[np.ndarray, np.ndarray]:
        """Sample an step_tuple from a gaussian step_tuple distribution defined by the means and log standard deviations."""
        actions = tf.random.normal(means.shape, means, tf.exp(log_stdevs))
        probabilities = self.log_probability(actions, means=means, log_stdevs=log_stdevs)

        return tf.reshape(actions, [-1]).numpy(), tf.squeeze(probabilities).numpy()

    def act_numpy(self, means: np.ndarray, log_stdevs: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """Sample an step_tuple from a gaussian step_tuple distribution defined by the means and log standard deviations."""
        actions = np.random.normal(means, np.exp(log_stdevs), size=means.shape)
        probabilities = self.log_probability_numpy(actions, means=means, log_stdevs=log_stdevs)

        return np.reshape(actions, [-1]), np.squeeze(probabilities)

    def sample(self, means: tf.Tensor, log_stdevs: tf.Tensor):
        """Sample from the Gaussian distribution."""
        action = tf.random.normal(means.shape, means, tf.exp(log_stdevs))
        return tf.reshape(action, [-1]).numpy()

    @tf.function
    def probability(self, samples: tf.Tensor, means: tf.Tensor, stdevs: tf.Tensor):
        """Calculate probability density for a given batch of potentially joint Gaussian PDF."""
        samples_transformed = (samples - means) / stdevs
        pdf = (tf.exp(-(tf.pow(samples_transformed, 2) / 2)) / tf.sqrt(2 * math.pi)) / stdevs
        return tf.math.reduce_prod(pdf, axis=-1)

    @staticmethod
    @tf.function
    def log_probability(samples: tf.Tensor, means: tf.Tensor, log_stdevs: tf.Tensor):
        """Calculate log probability density for a given batch of potentially joint Gaussian PDF.

        Input Shapes:
            - all: (B, A) or (B, S, A)

        Output Shapes:
            - all: (B) or (B, S)
        """
        log_likelihoods = (- tf.reduce_sum(log_stdevs, axis=-1)
                           - tf.math.log(2 * np.pi)
                           - (0.5 * tf.reduce_sum(tf.square(((samples - means) / tf.exp(log_stdevs))), axis=-1)))

        return log_likelihoods

    @staticmethod
    def log_probability_numpy(samples: np.ndarray, means: np.ndarray, log_stdevs: np.ndarray):
        """Calculate log probability density for a given batch of potentially joint Gaussian PDF.

        Input Shapes:
            - all: (B, A) or (B, S, A)

        Output Shapes:
            - all: (B) or (B, S)
        """
        log_likelihoods = (- np.sum(log_stdevs, axis=-1)
                           - np.log(2 * np.pi)
                           - (0.5 * np.sum(np.square(((samples - means) / np.exp(log_stdevs))), axis=-1)))

        return log_likelihoods

    @tf.function
    def _entropy_from_params(self, stdevs: tf.Tensor):
        """Calculate the joint entropy of Gaussian random variables described by their standard deviations.

        Input Shape: (B, A) or (B, S, A) for recurrent

        Since the given r.v.'s are independent, the sub-additivity property of entropy narrows down to an equality
        of the joint entropy and the sum of marginal entropies.
        """
        entropy = .5 * tf.math.log(2 * math.pi * math.e * tf.pow(stdevs, 2))
        return tf.reduce_sum(entropy, axis=-1)

    @tf.function
    def _entropy_from_log_params(self, log_stdevs: tf.Tensor):
        """Calculate the joint entropy of Gaussian random variables described by their log standard deviations.

        Input Shape: (B, A) or (B, S, A) for recurrent

        Since the given r.v.'serialization are independent, the subadditivity property of entropy narrows down to an equality
        of the joint entropy and the sum of marginal entropies.
        """
        entropy = .5 * (tf.math.log(math.pi * 2) + (tf.multiply(2.0, log_stdevs) + 1.0))
        return tf.reduce_sum(entropy, axis=-1)

    @tf.function
    def _approx_entropy_from_log(self, log_stdevs: tf.Tensor):
        """Calculate the joint entropy of Gaussian random variables described by their log standard deviations, but in
        an approximation. Essentially this removes any unnecessary scaling calculations and only leaves the bare
        log standard deviations as the entropy of the distribution.

        Input Shape: (B, A) or (B, S, A) for recurrent

        Since the given r.v.'serialization are independent, the subadditivity property of entropy narrows down to an equality
        of the joint entropy and the sum of marginal entropies.
        """
        return tf.reduce_sum(log_stdevs, axis=-1)

    @tf.function
    def entropy(self, params: Tuple[tf.Tensor, tf.Tensor]):
        """Calculate the joint entropy of Gaussian random variables described by their log standard deviations.

        Input Shape: (B, A) or (B, S, A) for recurrent

        Since the given r.v.'serialization are independent, the subadditivity property of entropy narrows down to an equality
        of the joint entropy and the sum of marginal entropies.
        """
        log_stdevs = params[1]
        return self._entropy_from_log_params(log_stdevs)

    def build_action_head(self, n_actions: Tuple[int], input_shape, batch_size, stdevs_from_latent=False):
        """Build a policy head for the gaussian distribution, for mean and stdev prediction."""
        n_actions = n_actions[0]

        inputs = tf.keras.Input(batch_shape=(batch_size,) + tuple(input_shape))

        means = tf.keras.layers.Dense(n_actions, name="means",
                                      kernel_initializer=tf.keras.initializers.Orthogonal(0.01),
                                      bias_initializer=tf.keras.initializers.Constant(0.0))(inputs)
        if stdevs_from_latent:
            stdevs = tf.keras.layers.Dense(n_actions, name="log_stdevs",
                                           kernel_initializer=tf.keras.initializers.Orthogonal(0.01),
                                           bias_initializer=tf.keras.initializers.Constant(0.0))(inputs)
        else:
            stdevs = StdevLayer(n_actions, name="log_stdevs")(means)

        return tf.keras.Model(inputs=inputs, outputs=[means, stdevs], name="gaussian_action_head")


class BetaPolicyDistribution(BaseContinuousPolicyDistribution):
    """Beta Distribution for bounded step_tuple spaces.

    The beta distribution has finite support in the interval [0, 1]. In contrast to infinitely supported distributions
    it is henceforth bias free and can be scaled to fit any bounded step_tuple space.
    """

    def __init__(self, env: gym.Env):
        super().__init__(env)

        # assure that rescaling is possible
        assert not np.any(np.isinf(self.action_min_values)) and not np.any(np.isinf(self.action_max_values))

    @property
    def short_name(self):
        """Policy's short identifier."""
        return "beta"

    @property
    def has_log_params(self):
        """Beta Distribution expects all parameters on standard scale."""
        return False

    def act(self, alphas: Union[tf.Tensor, np.ndarray], betas: Union[tf.Tensor, np.ndarray]):
        """Sample an action from a beta distribution."""
        actions = self.sample(alphas, betas)
        probabilities = tf.squeeze(self.log_probability(actions, alphas, betas)).numpy()

        return actions, probabilities

    def act_numpy(self, alphas: np.ndarray, betas: np.ndarray):
        """Sample an action from a beta distribution."""
        actions = np.random.beta(alphas, betas)

        # we need to prevent 0 and 1 as actions, otherwise log in probability calculation can fuck up
        actions = np.where(actions == 0, actions + EPSILON, actions)
        actions = np.where(actions == 1, actions - EPSILON, actions)

        actions = self._scale_sample_to_action_range_numpy(np.reshape(actions, [-1]))

        probabilities = tf.squeeze(self.log_probability_numpy(actions, alphas, betas))

        return actions, probabilities

    def act_deterministic(self, alphas: Union[tf.Tensor, np.ndarray], betas: Union[tf.Tensor, np.ndarray]):
        """Get action by deterministically taking the mode of the distribution."""
        actions = (alphas) / (alphas + betas)

        actions = self._scale_sample_to_action_range(np.reshape(actions, [-1])).numpy()

        return actions, self.log_probability(actions, alphas, betas)

    def sample(self, alphas: tf.Tensor, betas: tf.Tensor):
        """Sample from the Beta distribution."""
        actions = np.random.beta(alphas, betas).astype("float32")

        # we need to prevent 0 and 1 as actions, otherwise log in probability calculation can fuck up
        actions = np.where(actions == 0, actions + EPSILON, actions)
        actions = np.where(actions == 1, actions - EPSILON, actions)

        actions = self._scale_sample_to_action_range(np.reshape(actions, [-1])).numpy()

        return actions

    @tf.function
    def _scale_sample_to_action_range(self, sample) -> tf.Tensor:
        return tf.add(tf.multiply(sample, self.action_mm_diff), self.action_min_values)

    def _scale_sample_to_action_range_numpy(self, sample) -> np.ndarray:
        return np.add(np.multiply(sample, self.action_mm_diff), self.action_min_values)

    @tf.function
    def _scale_sample_to_distribution_range(self, sample) -> tf.Tensor:
        return tf.clip_by_value(tf.divide(tf.subtract(sample, self.action_min_values), self.action_mm_diff),
                                EPSILON,
                                1. - EPSILON)

    def _scale_sample_to_distribution_range_numpy(self, sample) -> np.ndarray:
        return np.divide(np.subtract(sample, self.action_min_values), self.action_mm_diff)

    @tf.function
    def probability(self, samples: tf.Tensor, alphas: tf.Tensor, betas: tf.Tensor):
        """Probability density of the Beta distribution."""
        samples = self._scale_sample_to_distribution_range(samples)

        top = tf.pow(samples, tf.subtract(alphas, 1.)) * tf.pow(tf.subtract(1., samples), tf.subtract(betas, 1.))
        bab = tf.multiply(tf.exp(tf.math.lgamma(alphas)),
                          tf.exp(tf.math.lgamma(betas)) / tf.exp(tf.math.lgamma(tf.add(alphas, betas))))

        return tf.math.reduce_prod(top / bab, axis=-1)

    @tf.function
    def log_probability(self, samples: tf.Tensor, alphas: tf.Tensor, betas: tf.Tensor):
        """Log probability utilizing the fact that tensorflow directly returns log of gamma function.

        Input Shape: (B, A)
        Output Shape: (B,)
        """
        rescaled_samples = self._scale_sample_to_distribution_range(samples)

        log_top = (tf.math.log(rescaled_samples)
                   * tf.subtract(alphas, 1.)
                   + tf.math.log(tf.subtract(1., rescaled_samples))
                   * tf.subtract(betas, 1.))
        log_pdf = log_top - tf.math.lgamma(alphas) - tf.math.lgamma(betas) + tf.math.lgamma(alphas + betas)

        return tf.math.reduce_sum(log_pdf, axis=-1)

    def log_probability_numpy(self, samples: np.ndarray, alphas: np.ndarray, betas: np.ndarray):
        """
        Input Shape: (B, A)
        Output Shape: (B,)
        """
        rescaled_samples = self._scale_sample_to_distribution_range_numpy(samples)
        log_pdf = scipy.stats.beta.logpdf(rescaled_samples, alphas, betas)

        return np.sum(log_pdf, axis=-1)

    @tf.function
    def entropy(self, params: Tuple[tf.Tensor, tf.Tensor]):
        """Entropy of the beta distribution."""
        return self._entropy_from_params(params)

    @tf.function
    def _entropy_from_params(self, params: Tuple[tf.Tensor, tf.Tensor]):
        """Entropy of the beta distribution"""
        alphas, betas = params

        # directly get log of bab to prevent numerical issues, lgamma is save because alpha beta always > 0
        bab_log = tf.math.lgamma(alphas) + (tf.math.lgamma(betas) - tf.math.lgamma(tf.add(alphas, betas)))

        # get all parts of equation; polygamma is save because alpha and beta cannot become 0
        a = tf.multiply(tf.subtract(alphas, 1.), tf.math.polygamma(0., alphas))
        b = tf.multiply(tf.subtract(betas, 1.), tf.math.polygamma(0., betas))
        ab = tf.multiply(tf.subtract(alphas + betas, 2.), tf.math.polygamma(0., tf.add(alphas, betas)))

        return tf.reduce_sum(bab_log - a - b + ab, axis=-1)

    def build_action_head(self, n_actions, input_shape, batch_size):
        """Build a policy head for the beta distribution, for alpha and beta prediction.

        """
        n_actions = n_actions[0]

        inputs = tf.keras.Input(batch_shape=(batch_size,) + tuple(input_shape))

        alphas = tf.keras.layers.Dense(n_actions, name="alphas", activation="softplus", dtype="float32",
                                       kernel_initializer=tf.keras.initializers.Orthogonal(1),
                                       bias_initializer=tf.keras.initializers.Constant(0.0))(inputs)

        betas = tf.keras.layers.Dense(n_actions, name="betas", activation="softplus", dtype="float32",
                                      kernel_initializer=tf.keras.initializers.Orthogonal(1),
                                      bias_initializer=tf.keras.initializers.Constant(0.0))(inputs)

        # guarantee a, b > 1 so that the distribution is concave and unimodal, see Chou, Maturana & Scherer (2017)
        alphas = tf.add(alphas, 1.)
        betas = tf.add(betas, 1.)

        return tf.keras.Model(inputs=inputs, outputs=[alphas, betas], name="beta_action_head")


class RBetaPolicyDistribution(BetaPolicyDistribution):
    """Beta Policy Distribution with input-independent entropy by reparametrization with mean and a+b."""

    @property
    def short_name(self):
        return "rbeta"

    def build_action_head(self, n_actions, input_shape, batch_size):
        """Build a policy head for the beta distribution, for alpha and beta prediction."""
        n_actions = n_actions[0]

        inputs = tf.keras.Input(batch_shape=(batch_size,) + tuple(input_shape))

        mu = tf.keras.layers.Dense(n_actions, name="mean", activation="sigmoid",
                                   kernel_initializer=tf.keras.initializers.Orthogonal(1),
                                   bias_initializer=tf.keras.initializers.Constant(0.0))(inputs)

        v = BetaDistributionSpreadLayer(n_actions, name="spread")(mu)

        # guarantee a, b > 1 by guaranteeing their sum to be 2
        v = tf.add(v, 2.)

        # derive alpha and beta
        alphas = tf.keras.layers.Add(name="alphas")([mu, v])
        betas = tf.keras.layers.Add(name="betas")([(-mu + 1), v])

        return tf.keras.Model(inputs=inputs, outputs=[alphas, betas], name="beta_action_head")


_distribution_short_name_map = {
    "gaussian": GaussianPolicyDistribution,
    "discrete": CategoricalPolicyDistribution,
    "categorical": CategoricalPolicyDistribution,
    "multi-categorical": MultiCategoricalPolicyDistribution,
    "multi-discrete": MultiCategoricalPolicyDistribution,
    "beta": BetaPolicyDistribution,
    "rbeta": RBetaPolicyDistribution
}


def get_distribution_by_short_name(name: str) -> type(BasePolicyDistribution):
    """Get a policy distribution object based on a short name identifier."""
    if name not in _distribution_short_name_map.keys():
        raise ValueError("Unknown distribution type.")

    return _distribution_short_name_map[name]



import numpy as np
import tensorflow as tf

from angorapy.utilities.model_utils import reset_states_masked

import pytest


def test_masked_state_reset():
    model = tf.keras.Sequential((
        tf.keras.layers.Dense(2, batch_input_shape=(7, None, 2)),
        tf.keras.layers.LSTM(5, stateful=True, name="larry", return_sequences=True),
        tf.keras.layers.LSTM(5, stateful=True, name="harry"))
    )

    l_layer = model.get_layer("larry")
    h_layer = model.get_layer("harry")
    l_layer.reset_states([s.numpy() + 9 for s in l_layer.states])
    h_layer.reset_states([s.numpy() + 9 for s in h_layer.states])

    reset_states_masked(
        [layer for layer in model.submodules if isinstance(layer, tf.keras.layers.RNN)],
        [True, True, False, True, False, False, True]
    )

    assert np.allclose([s.numpy() for s in model.get_layer("larry").states],
                       [s.numpy() for s in model.get_layer("harry").states])
    assert np.allclose([s.numpy() for s in model.get_layer("larry").states], [
        [0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0],
        [9, 9, 9, 9, 9],
        [0, 0, 0, 0, 0],
        [9, 9, 9, 9, 9],
        [9, 9, 9, 9, 9],
        [0, 0, 0, 0, 0],
    ])



#!/usr/bin/env python
"""Components that can be loaded into another network."""

import tensorflow as tf


def _build_fcn_component(input_dim: int, hidden_dim: int, output_dim: int, batch_size: int = None,
                         name: str = None):
    inputs = tf.keras.Input(batch_shape=(batch_size, input_dim))

    x = tf.keras.layers.Dense(hidden_dim)(inputs)
    x = tf.keras.layers.ReLU()(x)
    x = tf.keras.layers.Dense(output_dim)(x)
    x = tf.keras.layers.ReLU()(x)

    return tf.keras.Model(inputs=inputs, outputs=x, name=name)


def _build_encoding_sub_model(shape, batch_size, layer_sizes=(64, 64), name=None):
    assert len(layer_sizes) > 0, "You need at least one layer in an encoding submodel."

    inputs = tf.keras.Input(shape=shape, batch_size=batch_size, name=f"{name}_input")

    x = inputs
    for i in range(len(layer_sizes)):
        x = tf.keras.layers.Dense(layer_sizes[i],
                                  kernel_initializer=tf.keras.initializers.Orthogonal(gain=tf.sqrt(2.0)),
                                  bias_initializer=tf.constant_initializer(0.0),
                                  name=f"{name}_{i}")(x)
        x = tf.keras.layers.Activation("tanh")(x)

    return tf.keras.Model(inputs=inputs, outputs=x, name=name)



from typing import List, Dict, Tuple

import bokeh
import numpy
import numpy as np
import pandas as pd
import scipy
from bokeh import embed, layouts
from bokeh.io import export_svg
from bokeh.models import Span, Range1d, LinearAxis, ColumnDataSource, HoverTool
from bokeh.plotting import figure
from scipy import stats

from angorapy.utilities.monitor.plotting_base import palette, plot_styling, style_plot
from angorapy.utilities.statistics import mean_fill_nones


def plot_memory_usage(memory_trace: List):
    """Plot the development of memory used by the training."""
    x = list(range(len(memory_trace)))
    y = memory_trace

    p = figure(title="Memory Usage",
               x_axis_label='Cycle',
               y_axis_label='Memory in GB',
               y_range=(min(memory_trace), max(memory_trace)),
               **plot_styling)

    p.line(x, y, legend_label="RAM", line_width=2, color=palette[0])

    p.legend.location = "bottom_right"
    style_plot(p)

    return p


def plot_execution_times(cycle_timings, optimization_timings=None, gathering_timings=None):
    """Plot the execution times of a full cycle and optionally bot sub phases."""
    x = list(range(len(cycle_timings)))

    all_times = (cycle_timings
                 + (optimization_timings if optimization_timings is not None else [])
                 + (gathering_timings if gathering_timings is not None else []))

    if len(all_times) < 1:
        return None, None

    p = figure(title="Execution Times",
               x_axis_label='Cycle',
               y_axis_label='Seconds',
               y_range=(0, max(all_times)),
               **plot_styling)

    p.line(x, cycle_timings, legend_label="Full Cycle", line_width=2, color=palette[0])
    if optimization_timings is not None:
        p.line(x, optimization_timings, legend_label="Optimization Phase", line_width=2, color=palette[1])
    if gathering_timings is not None:
        p.line(x, gathering_timings, legend_label="Gathering Phase", line_width=2, color=palette[2])

    p.legend.location = "bottom_right"
    style_plot(p)

    return p


def plot_reward_progress(rewards: Dict[str, list], cycles_loaded, reward_threshold=None):
    """Plot the execution times of a full cycle and optionally bot sub phases."""
    means, stds = rewards["mean"], rewards["stdev"]
    stds = np.array(stds)
    stds[stds == None] = 0
    stds = stds * 0.2
    means = np.array(means)
    means = mean_fill_nones(means)

    x = list(range(len(means)))
    df = pd.DataFrame(data=dict(x=x, y=means, lower=np.subtract(means, stds), upper=np.add(means, stds)))
    value_range = max(df["upper"]) - min(df["lower"])

    tooltip = HoverTool(
        tooltips=[("Cycle", "@x"),
                  ("Reward", "@y")],
        mode="vline"
    )

    range_max = max(df["upper"]) + value_range * 0.2
    if reward_threshold is not None:
        range_max = max(range_max, reward_threshold * 1.1)
    p = figure(title="Average Rewards per Cycle",
               x_axis_label='Cycle',
               y_axis_label='Total Episode Return',
               y_range=(min(df["lower"]), range_max),
               x_range=(0, max(x)),
               **plot_styling)

    p.add_tools(tooltip)
    p.add_tools(bokeh.models.BoxZoomTool())
    p.add_tools(bokeh.models.ResetTool())
    p.add_tools(bokeh.models.SaveTool())

    # ERROR BAND
    error_band = bokeh.models.Band(
        base="x", lower="lower", upper="upper",
        source=ColumnDataSource(df.reset_index()),
        fill_color=palette[0],
        fill_alpha=0.2,
        line_color=palette[0],
        line_alpha=0.4,
    )
    p.add_layout(error_band)
    p.renderers.extend([error_band])

    # REWARD LINE
    p.line(x, means, legend_label="Reward", line_width=2, color=palette[0])

    # REWARD THRESHOLD
    if reward_threshold is not None:
        p.line(x, [reward_threshold for _ in x], line_color="green", line_width=2, line_alpha=0.7, line_dash="dashed",
               legend_label="Solution Threshold")

    # MAX VALUE MARKING
    x_max = np.argmax(means)
    y_max = np.max(means)
    p.add_layout(bokeh.models.Arrow(end=bokeh.models.NormalHead(size=10,
                                                                line_color="darkred",
                                                                line_width=2,
                                                                fill_color="red"),
                                    line_color="darkred",
                                    line_width=2,
                                    x_start=x_max, y_start=y_max + value_range * 0.07,
                                    x_end=x_max, y_end=y_max))
    p.add_layout(bokeh.models.Label(x=x_max, y=y_max + value_range * 0.075, text=str(y_max),
                                    border_line_color='black', border_line_alpha=0,
                                    background_fill_color='white', background_fill_alpha=1.0, text_align="center",
                                    text_line_height=1.5, text_font_size="10pt", text_color="darkred"

                                    ))

    load_markings = []
    for load_timing in cycles_loaded:
        load_markings.append(
            Span(location=load_timing[0], dimension="height", line_color="red", line_width=2, line_dash=[6, 3])
        )

    p.renderers.extend(load_markings)

    p.legend.location = "bottom_right"
    style_plot(p)

    return p


def plot_aux_perf_progress(aux_perf: Dict[str, list], cycles_loaded, perf_name: str="Performance Metric"):
    """Plot auxiliary performance measure."""
    perf_name = " ".join([n.capitalize() for n in perf_name.split("_")])

    if isinstance(aux_perf, list):
        means = aux_perf
        stds = np.zeros_like(aux_perf)
    else:
        means, stds = aux_perf["mean"], aux_perf["std"]
    stds = np.array(stds)
    stds[stds == None] = 0
    stds = stds * 0.2
    means = np.array(means)
    means = mean_fill_nones(means)

    x = list(range(len(means)))
    df = pd.DataFrame(data=dict(x=x, y=means, lower=np.subtract(means, stds), upper=np.add(means, stds)))
    value_range = max(df["upper"]) - min(df["lower"])

    tooltip = HoverTool(
        tooltips=[("Cycle", "@x"),
                  ("Reward", "@y")],
        mode="vline"
    )

    range_max = max(df["upper"]) + value_range * 0.2
    p = figure(title=f"{perf_name} per Cycle",
               x_axis_label='Cycle',
               y_axis_label=perf_name,
               y_range=(min(df["lower"]), range_max),
               x_range=(0, max(x)),
               **plot_styling)

    p.add_tools(tooltip)
    p.add_tools(bokeh.models.BoxZoomTool())
    p.add_tools(bokeh.models.ResetTool())
    p.add_tools(bokeh.models.SaveTool())

    # ERROR BAND
    error_band = bokeh.models.Band(
        base="x", lower="lower", upper="upper",
        source=ColumnDataSource(df.reset_index()),
        fill_color=palette[0],
        fill_alpha=0.2,
        line_color=palette[0],
        line_alpha=0.4,
    )
    p.add_layout(error_band)
    p.renderers.extend([error_band])

    # REWARD LINE
    p.line(x, means, legend_label="Reward", line_width=2, color=palette[0])

    # MAX VALUE MARKING
    x_max = np.argmax(means)
    y_max = np.max(means)
    p.add_layout(bokeh.models.Arrow(end=bokeh.models.NormalHead(size=10,
                                                                line_color="darkred",
                                                                line_width=2,
                                                                fill_color="red"),
                                    line_color="darkred",
                                    line_width=2,
                                    x_start=x_max, y_start=y_max + value_range * 0.07,
                                    x_end=x_max, y_end=y_max))
    p.add_layout(bokeh.models.Label(x=x_max, y=y_max + value_range * 0.075, text=str(y_max),
                                    border_line_color='black', border_line_alpha=0,
                                    background_fill_color='white', background_fill_alpha=1.0, text_align="center",
                                    text_line_height=1.5, text_font_size="10pt", text_color="darkred"

                                    ))

    load_markings = []
    for load_timing in cycles_loaded:
        load_markings.append(
            Span(location=load_timing[0], dimension="height", line_color="red", line_width=2, line_dash=[6, 3])
        )

    p.renderers.extend(load_markings)

    p.legend.location = "bottom_right"
    style_plot(p)

    return p


def compare_reward_progress(rewards: Dict[str, Dict[str, list]], reward_threshold=None):
    """Plot the execution times of a full cycle and optionally bot sub phases."""
    y_min = 0
    y_max = 0
    x_all = []

    p = figure(title="Average Rewards per Cycle",
               x_axis_label='Cycle',
               y_axis_label='Total Episode Return',
               **plot_styling)
    p.x_range.start = 0

    names_and_lines = []
    for i, (name, data) in enumerate(rewards.items()):
        means, stds = data["mean"], data["stdev"]
        stds = np.array(stds)
        stds[stds == None] = 0
        stds = stds * 0.2
        means = np.array(means)
        means = mean_fill_nones(means)

        y_min = min(min(means), y_min)
        y_max = max(max(means), y_max)

        x = list(range(len(means)))
        if len(x) > len(x_all):
            x_all = x

        df = pd.DataFrame(data=dict(x=x, y=means, lower=np.subtract(means, stds), upper=np.add(means, stds)))
        value_range = max(df["upper"]) - min(df["lower"])

        range_max = max(df["upper"]) + value_range * 0.1
        if reward_threshold is not None:
            range_max = max(range_max, reward_threshold * 1.1)

        # ERROR BAND
        error_band = bokeh.models.Band(
            base="x", lower="lower", upper="upper",
            source=ColumnDataSource(df.reset_index()),
            fill_color=palette[i % len(palette)],
            fill_alpha=0.2,
            line_color=palette[i % len(palette)],
            line_alpha=0.4,
        )
        p.add_layout(error_band)
        p.renderers.extend([error_band])

        # REWARD LINE
        reward_line = p.line(x, means, line_width=2, color=palette[i % len(palette)])
        names_and_lines.append((
            name, [reward_line]
        ))

    p.add_tools(bokeh.models.BoxZoomTool())
    p.add_tools(bokeh.models.ResetTool())
    p.add_tools(bokeh.models.SaveTool())

    # REWARD THRESHOLD
    if reward_threshold is not None:
        p.line(x_all, [reward_threshold for _ in x_all], line_color="green", line_width=2, line_alpha=0.7, line_dash="dashed",
               legend_label="Solution Threshold")

    p.y_range = bokeh.models.Range1d(y_min, y_max)

    legend = bokeh.models.Legend(items=names_and_lines, location="bottom_center")
    p.add_layout(legend, "below")

    style_plot(p)
    p.height = 750 + legend.glyph_height * len(rewards)

    return p


def make_grouped_reward_progress_figure(
        rewards: Dict[str, Tuple[int, Dict[str, list]]],
        reward_threshold,
        title=None,
        no_y_label=False
):
    y_min = 0
    y_max = 0 if reward_threshold is None else reward_threshold
    x_all = []

    p = figure(title="Average Rewards per Cycle" if title is None else title,
               x_axis_label='Cycle',
               y_axis_label='Total Episode Return' if not no_y_label else "",
               **plot_styling)
    p.x_range.start = 0

    by_groups = {}
    for name, (group, data) in rewards.items():
        if group not in by_groups.keys():
            by_groups[group] = []

        by_groups[group].append(data)

    i = 0
    names_and_lines = []
    tooltip_parts = []
    for group, datas in by_groups.items():
        means = np.mean([mean_fill_nones(d["mean"]) for d in datas], axis=0)
        stds = np.std([d["mean"] for d in datas], axis=0)

        y_min = min(min(means), y_min)
        y_max = max(max(means + stds), y_max)

        x = list(range(len(means)))
        if len(x) > len(x_all):
            x_all = x

        df = pd.DataFrame(data=dict(x=x, y=means, lower=np.subtract(means, stds), upper=np.add(means, stds)))
        value_range = max(df["upper"]) - min(df["lower"])

        tooltip_parts.append((f"Group {group}", "@y"))

        range_max = max(df["upper"]) + value_range * 0.1

        # ERROR BAND
        error_band = bokeh.models.Band(
            base="x", lower="lower", upper="upper",
            source=ColumnDataSource(df.reset_index()),
            fill_color=palette[i % len(palette)],
            fill_alpha=0.2,
            line_color=palette[i % len(palette)],
            line_alpha=0.4,
        )
        p.add_layout(error_band)
        p.renderers.extend([error_band])

        # REWARD LINE
        reward_line = p.line(x, means, line_width=2, color=palette[i % len(palette)])
        names_and_lines.append((
            str(group), [reward_line]
        ))

        i += 1

    tooltip_parts.append(("Cycle", "@x"))
    tooltip = HoverTool(
        tooltips=tooltip_parts,
        mode="vline"
    )

    p.add_tools(tooltip)

    p.add_tools(bokeh.models.BoxZoomTool())
    p.add_tools(bokeh.models.ResetTool())
    p.add_tools(bokeh.models.SaveTool())

    # REWARD THRESHOLD
    if reward_threshold is not None:
        p.line(x_all, [reward_threshold for _ in x_all], line_color="green", line_width=2, line_alpha=0.7,
               line_dash="dashed",
               legend_label="Solution Threshold")

    p.y_range = bokeh.models.Range1d(y_min, y_max)

    legend = bokeh.models.Legend(items=names_and_lines, location="bottom_center", orientation="horizontal")
    p.add_layout(legend, "below")

    style_plot(p)

    return p


def grouped_reward_progress(rewards: Dict[str, Tuple[int, Dict[str, list]]], reward_threshold=None):
    p = make_grouped_reward_progress_figure(rewards, reward_threshold)
    p.height = 750

    return p


def group_preview(groups: Dict[str, Dict[str, list]], alt_titles=None, reward_thresholds=None):
    figures = []
    for i, group in enumerate(groups.keys()):
        ids_with_rewards = groups[group]

        p = make_grouped_reward_progress_figure(
            {id: ("", progress["rewards"]) for id, progress in ids_with_rewards.items()},
            None if reward_thresholds is None else reward_thresholds[group],
            title=group if alt_titles is None else alt_titles[group],
            no_y_label=(i != 0)
        )

        p.legend.visible = False

        figures.append(p)

    gp = layouts.gridplot(figures, ncols=4, sizing_mode="scale_width")

    return gp


def plot_length_progress(lengths: Dict[str, list], cycles_loaded):
    """Plot development of the lengths of episodes throughout training."""
    means, stds = lengths["mean"], lengths["stdev"]
    stds = np.array(stds)
    stds[stds == None] = 0
    stds = stds * 0.2
    means = np.array(means)
    means[means == None] = 0

    x = list(range(len(means)))
    df = pd.DataFrame(data=dict(x=x, y=means, lower=np.subtract(means, stds), upper=np.add(means, stds)))
    value_range = max(df["upper"]) - min(df["lower"])

    tooltip = HoverTool(
        tooltips=[("Cycle", "@x"),
                  ("Length", "@y")],
        mode="vline"
    )

    p = figure(title="Average Length of Episodes",
               x_axis_label='Cycle',
               y_axis_label='Total Episode Return',
               y_range=(min(df["lower"]), max(df["upper"]) + value_range * 0.1),
               x_range=(0, max(x)),
               **plot_styling)

    p.add_tools(tooltip)
    p.add_tools(bokeh.models.BoxZoomTool())
    p.add_tools(bokeh.models.ResetTool())
    p.add_tools(bokeh.models.SaveTool())

    # ERROR BAND
    error_band = bokeh.models.Band(
        base="x", lower="lower", upper="upper",
        source=ColumnDataSource(df.reset_index()),
        fill_color=palette[1],
        fill_alpha=0.2,
        line_color=palette[1],
        line_alpha=0.4,
    )
    p.add_layout(error_band)
    p.renderers.extend([error_band])

    # REWARD LINE
    p.line(x, means, legend_label="Episode Length (time steps)", line_width=2, color=palette[1])

    # MAX VALUE MARKING
    x_max = np.argmax(means)
    y_max = np.max(means)
    p.add_layout(bokeh.models.Arrow(end=bokeh.models.NormalHead(size=15,
                                                                line_color="darkred",
                                                                line_width=2,
                                                                fill_color="red"),
                                    line_color="darkred",
                                    line_width=2,
                                    x_start=x_max, y_start=y_max + value_range * 0.1,
                                    x_end=x_max, y_end=y_max))
    p.add_layout(bokeh.models.Label(x=x_max, y=y_max + value_range * 0.1, text=str(y_max),
                                    border_line_color='black', border_line_alpha=1.0,
                                    background_fill_color='white', background_fill_alpha=1.0, text_align="center",
                                    text_line_height=1.5, text_font_size="10pt",

                                    ))

    load_markings = []
    for load_timing in cycles_loaded:
        load_markings.append(
            Span(location=load_timing[0], dimension="height", line_color="red", line_width=2, line_dash=[6, 3])
        )

    p.renderers.extend(load_markings)

    p.legend.location = "bottom_right"
    style_plot(p)

    return p


def plot_distribution(metric: List[float], name="Metric", color=0) -> Tuple:
    hist, edges = np.histogram(metric, density=True, bins=50)

    p = figure(title=f"Distribution of {name}",
               x_axis_label=name,
               y_range=(0, np.max(hist)),
               **plot_styling)

    if len(metric) == 0:
        return p

    width = (max(metric) - min(metric)) / len(hist) * 0.9
    hist_glyph = p.rect(x=edges, y=hist / 2, height=hist, width=width, line_color="white", fill_color=palette[color])

    x, mean, sigma = np.linspace(min(metric), max(metric), 50), np.mean(metric), np.std(metric)

    tooltips = [("Reward", "@x")]
    try:
        pdf = stats.gaussian_kde(metric, bw_method=0.4).pdf(x)
        p.line(x, pdf, line_color=palette[color], line_width=4, alpha=0.7, )
        tooltips.append(("Density", "@y"))
    except np.linalg.LinAlgError as e:
        pass

    tooltip = HoverTool(
        tooltips=tooltips,
        mode="vline",
        renderers=[hist_glyph]
    )
    p.add_tools(tooltip)
    p.add_tools(bokeh.models.SaveTool())

    p.legend.location = "bottom_right"
    style_plot(p)

    return p


def plot_loss(loss, rewards, name, color_id=0):
    """Plot a loss as it develops over cycles."""
    x = list(range(len(loss)))
    loss = np.array(loss, dtype=float)
    rewards = np.array(rewards, dtype=float)

    loss[np.isnan(loss)] = None

    p = figure(title=name,
               x_axis_label='Cycle',
               y_axis_label='Loss',
               y_range=(min(loss), max(loss)),
               **plot_styling)

    p.extra_y_ranges = {"Reward": Range1d(start=min(rewards), end=max(rewards))}
    p.add_layout(LinearAxis(y_range_name="Reward"), "right")

    p.line(x, rewards, legend_label="Reward", line_width=2, color="lightgrey", y_range_name="Reward")
    p.line(x, loss, legend_label=name, line_width=2, color=palette[color_id])

    p.add_tools(bokeh.models.SaveTool())


    p.legend.location = "bottom_right"
    style_plot(p)

    return p


def plot_preprocessor(preprocessor_data: Dict[str, List[Dict[str, float]]]):
    """Plot progression of preprocessor means."""
    x = list(range(len(preprocessor_data["mean"])))

    plots = []
    for i, sense in enumerate(preprocessor_data["mean"][0].keys()):
        p = figure(title=f"{sense.capitalize()}",
                   x_axis_label='Cycle',
                   y_axis_label='Running Mean',
                   **plot_styling)

        trace = [point[sense] for point in preprocessor_data["mean"]]
        if len(trace) > 1:
            trace = trace[1:]

            if i == 0:
                x = x[1:]

        p.line(x, trace, legend_label=str(sense), line_width=2, color=palette[i])

        p.add_tools(bokeh.models.SaveTool())

        p.legend.location = "bottom_right"
        style_plot(p)

        plots.append(p)

    p = bokeh.layouts.row(plots)
    p.sizing_mode = "stretch_width"
    return p



from typing import Dict

import tensorflow as tf

from angorapy.agent.ppo.optim import learn_on_batch
from angorapy.utilities.model_utils import reset_states_masked_tf
from angorapy.utilities.core import detect_finished_episodes


def ff_train_step(batch,
                  joint,
                  distribution,
                  continuous_control,
                  clip_values,
                  gradient_clipping,
                  clip,
                  c_value,
                  c_entropy,
                  is_recurrent,
                  optimizer):
    grad, ent, pi_loss, v_loss = learn_on_batch(
        batch=batch, joint=joint, distribution=distribution,
        continuous_control=continuous_control, clip_values=clip_values,
        gradient_clipping=gradient_clipping, clipping_bound=clip, c_value=c_value,
        c_entropy=c_entropy, is_recurrent=is_recurrent)
    optimizer.apply_gradients(zip(grad, joint.trainable_variables))

    return ent, pi_loss, v_loss


@tf.function
def _split_batch(batch: Dict[str, tf.Tensor], n_chunks_per_trajectory_per_batch: int):
    _split_part = lambda bv: tf.reshape(bv,
                                       [bv.shape[0]]
                                       + [bv.shape[1] // n_chunks_per_trajectory_per_batch,
                                          n_chunks_per_trajectory_per_batch]
                                       + bv.shape[2:])

    return {k: _split_part(v) for k, v in batch.items()}


@tf.function
def _do_batch_op(joint,
                 recurrent_layers,
                 n_chunks_per_trajectory_per_batch,
                 batch_i,
                 split_batch,
                 distribution,
                 continuous_control,
                 clip_values,
                 gradient_clipping,
                 clip,
                 c_value,
                 c_entropy,
                 is_recurrent):
    batch_grad = [tf.zeros_like(tv) for tv in joint.trainable_variables]
    batch_ent = tf.constant(0.)
    batch_pi_loss = tf.constant(0.)
    batch_v_loss = tf.constant(0.)

    for chunk_i in tf.range(n_chunks_per_trajectory_per_batch):
        partial_batch = {k: v[:, batch_i, chunk_i, ...] for k, v in split_batch.items()}
        grad, ent, pi_loss, v_loss = learn_on_batch(
            batch=partial_batch, joint=joint, distribution=distribution,
            continuous_control=continuous_control, clip_values=clip_values,
            gradient_clipping=gradient_clipping, clipping_bound=clip,
            c_value=c_value, c_entropy=c_entropy, is_recurrent=is_recurrent)

        batch_grad = tf.nest.map_structure(lambda bg, g: tf.add(bg, g), batch_grad, grad)
        reset_mask = detect_finished_episodes(partial_batch["done"])

        batch_ent += ent
        batch_pi_loss += pi_loss
        batch_v_loss += v_loss

        # make partial RNN state resets
        reset_states_masked_tf(recurrent_layers, reset_mask)

    batch_grad = tf.nest.map_structure(lambda b: b / n_chunks_per_trajectory_per_batch, batch_grad)

    return batch_grad, batch_ent, batch_pi_loss, batch_v_loss


def recurrent_train_step(super_batch: dict,
                         batch_size: tf.int32,
                         joint,
                         distribution,
                         continuous_control,
                         clip_values,
                         gradient_clipping,
                         clip,
                         c_value,
                         c_entropy,
                         is_recurrent,
                         optimizer,
                         pbar=None):
    """Recurrent train step, using truncated back propagation through time.

    Incoming batch shape: (BATCH_SIZE, N_SUBSEQUENCES, SUBSEQUENCE_LENGTH, *STATE_DIMS)
    We split along the N_SUBSEQUENCES dimension to get batches of single subsequences that can be fed into the model
    chronologically to adhere to statefulness if the following line throws a CPU to GPU error this is most likely due
    to too little memory on the GPU; lower the number of worker/horizon/... ."""
    n_trajectories_per_batch, n_chunks_per_trajectory_per_batch = batch_size

    recurrent_layers = [layer for layer in joint.submodules if isinstance(layer, tf.keras.layers.RNN)]

    batch_ent = tf.constant(0.)
    batch_pi_loss = tf.constant(0.)
    batch_v_loss = tf.constant(0.)

    batch_ents = []
    batch_pi_losses = []
    batch_v_losses = []

    split_batch = _split_batch(super_batch, n_chunks_per_trajectory_per_batch)
    for batch_i in tf.range(split_batch["advantage"].shape[1]):
        batch_grad, batch_ent, batch_pi_loss, batch_v_loss = _do_batch_op(
            joint,
            recurrent_layers,
            n_chunks_per_trajectory_per_batch,
            batch_i,
            split_batch,
            distribution,
            continuous_control,
            clip_values,
            gradient_clipping,
            clip,
            c_value,
            c_entropy,
            is_recurrent
        )
        optimizer.apply_gradients(zip(batch_grad, joint.trainable_variables))
        batch_ents.append(batch_ent)
        batch_pi_losses.append(batch_pi_loss)
        batch_v_losses.append(batch_v_loss)

        if pbar is not None:
            pbar.update(1)

    return tf.reduce_mean(batch_ents), tf.reduce_mean(batch_pi_losses), tf.reduce_mean(batch_v_losses)



import itertools

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
import statsmodels.api as sm
from matplotlib import colors
from tqdm import tqdm

from angorapy import register_model
from angorapy.agent import PPOAgent
from dexterity.model import build_fpn_models, build_fpn_v2_models, build_fpn_v3_models

register_model(build_fpn_models)
register_model(build_fpn_v2_models)
register_model(build_fpn_v3_models)


def get_timeseries_data(agent, random_steps=True):
    env = agent.env
    if random_steps:
        agent.policy, agent.value, agent.joint = agent.build_models(
            agent.joint.get_weights(),
            batch_size=1,
            sequence_length=1
        )

    state_trace = []
    state, info = env.reset()
    for t in range(100):
        if not random_steps:
            action, _ = agent.distribution.act(*agent.policy(state.with_leading_dims(time=True).dict_as_tf()))
        else:
            action = env.action_space.sample()
        state, r, terminated, truncated, i = env.step(np.atleast_1d(action))
        state_trace.append(np.concatenate([
            state.dict()["proprioception"],
            state.dict()["vision"],
            # state.dict()["touch"] + np.random.normal(0, 0.01, state.dict()["touch"].shape),
            state.dict()["asymmetric"],
        ]))

        if terminated:
            state, info = env.reset()

    data = np.diff(np.array(state_trace), 0)

    return data


def calculate_timewise_autocorrelation(data, n_lags=30):
    """Calculate autocorrelations between timeseries at all lags up to n_lags."""
    total_corr = []
    for i_var in range(data.shape[1]):
        timeseries_i = data[:, i_var]
        total_corr.append(np.abs(sm.tsa.acf(timeseries_i, nlags=n_lags - 1)))

    return np.array(total_corr).T


def calculate_timewise_crosscorrelations(data, n_lags=30):
    """Calculate autocorrelations between timeseries at all lags up to n_lags."""
    total_corr = [[] for i in range(n_lags)]
    for i_var in range(data.shape[1]):
        timeseries_i = data[:, i_var]

        for j_var in range(i_var, data.shape[1]):
            timeseries_j = data[:, j_var]
            for lag in range(n_lags):
                corr = np.corrcoef(timeseries_i, np.roll(timeseries_j, lag))[0, 1]
                total_corr[lag].append(np.abs(corr))

    return total_corr


def calculate_statewise_correlation(data, n_lags=30):
    corr_coeffs = np.corrcoef(data)
    # corr_coeffs = ((data @ data.T) / (np.linalg.norm(data, axis=1) ** 2))
    lag_shifted_corr_coeffs = np.zeros_like(corr_coeffs)
    for row in range(n_lags):
        lag_shifted_corr_coeffs[row] = np.roll(corr_coeffs[row], -row)

    return lag_shifted_corr_coeffs[:n_lags, :n_lags]


if __name__ == '__main__':
    # EXPERIMENTAL PARAMETERS
    n_samples = 15
    n_lags = 30

    # PLOT PARAMETERS
    MARKER_SIZE = 4

    violin_ax = plt.subplot2grid((3, 3), (0, 0), colspan=3, title="Pairwise Cross-correlations")
    acf_ax1 = plt.subplot2grid((3, 3), (1, 0), colspan=1, title="Pairwise Cross-correlations")
    acf_ax2 = plt.subplot2grid((3, 3), (1, 1), colspan=1, title="Autocorrelations")
    acf_ax3 = plt.subplot2grid((3, 3), (1, 2), colspan=1, title="State Correlations")
    acf_ax4 = plt.subplot2grid((3, 3), (2, 0), colspan=1)
    acf_ax5 = plt.subplot2grid((3, 3), (2, 1), colspan=1)
    acf_ax6 = plt.subplot2grid((3, 3), (2, 2), colspan=1)

    agent = PPOAgent.from_agent_state(1692396321151529, from_iteration="best", path_modifier="../../../")
    autocorrelations, crosscorrelations, statewise_correlations = [], [], []
    for i in tqdm(range(n_samples)):
        data = get_timeseries_data(agent)
        data = np.array(data) - np.mean(data, axis=0)

        autocorrelations.append(calculate_timewise_autocorrelation(data, n_lags=n_lags + 1))
        crosscorrelations.append(calculate_timewise_crosscorrelations(data, n_lags=n_lags + 1))
        statewise_correlations.append(calculate_statewise_correlation(data, n_lags=n_lags + 1))

    # stack samples
    sample_mean_autocorrelation = np.mean(np.stack(autocorrelations, 0), 0)
    sample_mean_crosscorrelation = np.mean(np.stack(crosscorrelations, 0), 0)
    sample_mean_statewise_correlation = np.mean(np.stack(statewise_correlations, 0), 0)

    # make crosscorrelation violin plot
    lags = (np.ones_like(sample_mean_crosscorrelation) * np.expand_dims(
        np.arange(sample_mean_crosscorrelation.shape[0]), [1])).flatten().astype(int)
    df = pd.DataFrame({"Lag": lags, "Correlation": np.array(sample_mean_crosscorrelation).flatten()})

    plt.gcf().set_size_inches(16, 8)

    # violoinplot with all violins filled with color lightblue
    sns.violinplot(
        data=df,
        x="Lag",
        y="Correlation",
        ax=violin_ax,
        linecolor="tab:blue",
        color="skyblue",
        saturation=1,
        inner_kws=dict(
            box_width=3,
            whis_width=0,
            markerfacecolor="black",
            markeredgecolor="black",
        )
    )

    # make auto and crosscorrelation plots
    ax_i = 0
    acf_axes = [acf_ax1, acf_ax4, acf_ax2, acf_ax5]
    for a_or_c_corr, combine_by in itertools.product([sample_mean_crosscorrelation, sample_mean_autocorrelation],
                                                     ["mean", "max"]):
        if combine_by == "mean":
            variable_mean_correlations = np.mean(a_or_c_corr, axis=-1)

            variable_se_correlations = (
                    np.std(a_or_c_corr, axis=-1)
                    / np.sqrt(a_or_c_corr.shape[-1])
            )

            acf_axes[ax_i].fill_between(
                range(n_lags + 1),
                variable_mean_correlations - variable_se_correlations,
                variable_mean_correlations + variable_se_correlations,
                color="skyblue"
            )
        elif combine_by == "max":
            variable_mean_correlations = np.max(a_or_c_corr, axis=-1)
        else:
            raise ValueError(f"combine_by must be 'mean' or 'max', not {combine_by}")

        markerline, stemlines, baseline = acf_axes[ax_i].stem(range(n_lags + 1), variable_mean_correlations)
        markerline.set_markersize(MARKER_SIZE)

        acf_axes[ax_i].vlines(
            16,
            ymin=0,
            ymax=1,
            linestyles="dashed",
            color="red"
        )

        acf_axes[ax_i].set_xlabel("Lag")
        acf_axes[ax_i].set_ylabel(f"{combine_by.capitalize()} Absolute Correlation")
        acf_axes[ax_i].set_ylim(0, 1.1)

        ax_i += 1

    # make statewise correlation plot
    variable_mean_correlations = np.mean(sample_mean_statewise_correlation, axis=-2)
    variable_ci_statewise_correlation = (
            np.std(sample_mean_statewise_correlation, axis=-2)
            / np.sqrt(sample_mean_statewise_correlation.shape[-2])
    )

    markerline, stemlines, baseline = acf_ax6.stem(range(n_lags + 1), variable_mean_correlations)
    markerline.set_markersize(MARKER_SIZE)
    acf_ax6.fill_between(
        range(n_lags + 1),
        variable_mean_correlations - variable_ci_statewise_correlation,
        variable_mean_correlations + variable_ci_statewise_correlation,
        color="skyblue"
    )
    acf_ax6.set_xlabel("Lag")
    acf_ax6.set_ylabel("Mean Correlation")

    acf_ax6.vlines(
        16,
        ymin=0,
        ymax=1,
        linestyles="dashed",
        color="red"
    )

    # make statewise correlation plot
    variable_mean_correlations = np.mean(np.abs(sample_mean_statewise_correlation), axis=-2)
    variable_ci_statewise_correlation = (
            np.std(np.abs(sample_mean_statewise_correlation), axis=-2)
            / np.sqrt(sample_mean_statewise_correlation.shape[-2])
    )

    markerline, stemlines, baseline = acf_ax3.stem(range(n_lags + 1), variable_mean_correlations)
    markerline.set_markersize(MARKER_SIZE)
    acf_ax3.fill_between(
        range(n_lags + 1),
        variable_mean_correlations - variable_ci_statewise_correlation,
        variable_mean_correlations + variable_ci_statewise_correlation,
        color="skyblue"
    )
    acf_ax3.set_xlabel("Lag")
    acf_ax3.set_ylabel("Mean Absolute Correlation")
    acf_ax3.set_ylim(0, 1.1)

    acf_ax3.vlines(
        16,
        ymin=0,
        ymax=1,
        linestyles="dashed",
        color="red"
    )

    plt.tight_layout()

    plt.savefig("../../../docs/figures/combined-correlation-plot.pdf", format="pdf", bbox_inches="tight")
    plt.show()



import os

import gymnasium as gym
import keras_cortex
import tensorflow as tf
from tensorflow.keras.layers import TimeDistributed as TD
# from tensorflow_core.python.keras.utils import plot_model
from tensorflow.keras.utils import plot_model

from angorapy.common.const import VISION_WH
from angorapy.common.policies import BasePolicyDistribution, MultiCategoricalPolicyDistribution
from angorapy.models import register_model
from angorapy.utilities.core import env_extract_dims


class OrthogonalMP(tf.keras.initializers.Orthogonal):

    def __call__(self, shape, dtype=None):
        return super().__call__(shape, dtype=dtype)


def build_ssc_module(batch_and_sequence_shape, touch_input_shape, activation=tf.keras.layers.ReLU):
    """Build the model of the SSC."""

    ssc_input = tf.keras.Input(batch_shape=batch_and_sequence_shape + touch_input_shape, name="SSCInput")
    ssc = TD(tf.keras.layers.Dense(296, name="SSC_1"), name="TD_ssc_1")(ssc_input)
    ssc = activation(name="SSC_activation_1")(ssc)
    ssc = TD(tf.keras.layers.Dense(296, name="SSC_2"), name="TD_ssc_2")(ssc)
    ssc = activation(name="SSC_activation_2")(ssc)

    return tf.keras.Model(inputs=ssc_input, outputs=ssc, name="SomatosensoryCortex")


def build_ppc_module(batch_and_sequence_shape, vc_input_shape, ssc_input_shape, activation=tf.keras.layers.ReLU):
    """Build the PPC model."""
    vc_input = tf.keras.Input(batch_shape=batch_and_sequence_shape + vc_input_shape, name="VCInput")
    ssc_input = tf.keras.Input(batch_shape=batch_and_sequence_shape + ssc_input_shape, name="SSCInput")

    spl = TD(tf.keras.layers.Dense(456, name="SPL"), name="TD_SPL")(ssc_input)
    spl = activation(name="SPL_activation")(spl)

    ipl_input = tf.keras.layers.concatenate([ssc_input, vc_input, spl])
    ipl = TD(tf.keras.layers.Dense(672, name="IPL"), name="TD_IPL")(ipl_input)
    ipl = activation(name="IPL_activation")(ipl)

    ips_input = tf.keras.layers.concatenate([ipl, ssc_input, vc_input, spl, ssc_input])
    ips = TD(tf.keras.layers.Dense(344, name="IPS"), name="TD_IPS")(ips_input)
    ips = activation(name="IPS_activation")(ips)

    return tf.keras.Model(inputs=[vc_input, ssc_input], outputs=[spl, ipl, ips], name="PosteriorParietalCortex")


def build_pfc_module(batch_and_sequence_shape, goal_input_shape, ipl_input_shape, ips_input_shape, activation=tf.keras.layers.ReLU):
    """Build the PFC model. Returns output of MCC, LPFC."""

    goal_input = tf.keras.Input(batch_shape=batch_and_sequence_shape + goal_input_shape, name="Goal Input")
    ipl_input = tf.keras.Input(batch_shape=batch_and_sequence_shape + ipl_input_shape, name="IPL Input")
    ips_input = tf.keras.Input(batch_shape=batch_and_sequence_shape + ips_input_shape, name="IPS Input")

    mcc = TD(tf.keras.layers.Dense(216, name="MCC"), name="TD_mcc")(goal_input)
    mcc = activation(name="MCC_activation")(mcc)

    lpfc_input = tf.keras.layers.concatenate([mcc, ipl_input, ips_input])
    lpfc = TD(tf.keras.layers.Dense(360, name="LPFC"), name="TD_lpfc")(lpfc_input)
    lpfc = activation(name="LPFC_activation")(lpfc)

    return tf.keras.Model(inputs=[goal_input, ipl_input, ips_input], outputs=[mcc, lpfc], name="PrefrontalCortex")


def build_mc_module(batch_and_sequence_shape, lpfc_input_shape, spl_input_shape,
                    ssc_input_shape, rnn_class=tf.keras.layers.LSTM, activation=tf.keras.layers.ReLU):
    lpfc_input = tf.keras.Input(batch_shape=batch_and_sequence_shape + lpfc_input_shape, name="LPFCInput")
    spl_input = tf.keras.Input(batch_shape=batch_and_sequence_shape + spl_input_shape, name="IPLInput")
    ssc_input = tf.keras.Input(batch_shape=batch_and_sequence_shape + ssc_input_shape, name="SSCInput")

    pmc_input = tf.keras.layers.concatenate([lpfc_input, spl_input, ssc_input])

    pmc, *_ = rnn_class(256,
                        stateful=True,
                        return_sequences=True,
                        batch_size=batch_and_sequence_shape[0],
                        return_state=True,
                        name="pmc_recurrent_layer",
                        recurrent_initializer=OrthogonalMP(),
                        dtype="float32",

                        )(pmc_input)

    m1_input = tf.keras.layers.concatenate([pmc, ssc_input, spl_input])

    m1 = TD(tf.keras.layers.Dense(96, name="m1"), )(m1_input)
    m1 = activation(name="M1_activation")(m1)

    return tf.keras.Model(inputs=[lpfc_input, spl_input, ssc_input],
                          outputs=[pmc, m1], name="MotorCortex")


@register_model("shadowv2")
def build_shadow_v2_brain_base(env: gym.Env, distribution: BasePolicyDistribution, bs: int = 1, model_type: str = "rnn",
                            blind: bool = False, sequence_length=1, activation=tf.keras.layers.ReLU, **kwargs):
    """Build network for the shadow hand task, version 2."""
    state_dimensionality, n_actions = env_extract_dims(env)

    rnn_choice = {"rnn": tf.keras.layers.SimpleRNN,
                  "lstm": tf.keras.layers.LSTM,
                  "gru": tf.keras.layers.GRU}[model_type]

    # inputs
    visual_input = tf.keras.Input(batch_shape=(bs, sequence_length, *state_dimensionality["vision"],), name="vision")
    proprio_in = tf.keras.Input(batch_shape=(bs, sequence_length, *state_dimensionality["proprioception"],),
                                name="proprioception")
    touch_in = tf.keras.Input(batch_shape=(bs, sequence_length, *state_dimensionality["touch"],),
                              name="touch")
    goal_input = tf.keras.Input(batch_shape=(bs, sequence_length, *state_dimensionality["goal"],), name="goal")

    input_list = [visual_input, proprio_in, touch_in, goal_input]

    touch = tf.keras.layers.concatenate([proprio_in, touch_in])
    touch_masked = tf.keras.layers.Masking(
        batch_input_shape=touch.shape
    )(touch)

    if blind:
        vc = visual_input
    else:
        visual_component = keras_cortex.cornet.CORNetZ(7)
        vc = TD(visual_component, name="visual_component")(visual_input)

    vision_masked = tf.keras.layers.Masking(batch_input_shape=vc.shape)(vc)
    goal_masked = tf.keras.layers.Masking(batch_input_shape=goal_input.shape)(goal_input)

    ssc = build_ssc_module((bs, sequence_length,),
                           (touch_masked.shape[-1],),
                           activation=activation)
    ssc_out = ssc(touch_masked)

    ppc = build_ppc_module((bs, sequence_length,),
                           (vision_masked.shape[-1],),
                           (ssc.output_shape[-1],),
                           activation=activation)
    spl_out, ipl_out, ips_out = ppc([vision_masked, ssc_out])

    pfc = build_pfc_module((bs, sequence_length,),
                           (goal_masked.shape[-1],),
                           (ipl_out.shape[-1],),
                           (ips_out.shape[-1],),
                           activation=activation)
    mcc_out, lpfc_out = pfc([goal_masked, ipl_out, ips_out])

    mc = build_mc_module((bs, sequence_length,),
                         (lpfc_out.shape[-1],),
                         (spl_out.shape[-1],),
                         (ssc_out.shape[-1],), rnn_class=rnn_choice,
                         activation=activation)
    pmc_out, m1_out = mc([lpfc_out, spl_out, ssc_out])

    # policy head
    policy_out = distribution.build_action_head(n_actions, m1_out.shape[1:], bs)(m1_out)

    # value head
    value_inputs = [pmc_out, mcc_out, ips_out, lpfc_out, ssc_out]
    if "asymmetric" in state_dimensionality.keys():
        asymmetric = tf.keras.Input(batch_shape=(bs, sequence_length,) + state_dimensionality["asymmetric"],
                                      name="asymmetric")
        input_list.append(asymmetric)
        value_inputs.append(asymmetric)

    value_in = tf.keras.layers.concatenate(value_inputs)
    value_out = tf.keras.layers.Dense(512)(value_in)
    value_out = activation()(value_out)
    value_out = tf.keras.layers.Dense(512)(value_out)
    value_out = activation()(value_out)
    value_out = tf.keras.layers.Dense(1, name="value", dtype="float32")(value_out)

    # define models
    policy = tf.keras.Model(inputs=input_list, outputs=[policy_out], name="shadow_brain_policy")
    value = tf.keras.Model(inputs=input_list, outputs=[value_out], name="shadow_brain_value")
    joint = tf.keras.Model(inputs=input_list, outputs=[policy_out, value_out],
                           name="shadow_v2" + ("_visual" if not blind else ""))

    return policy, value, joint


if __name__ == "__main__":
    from angorapy import make_task

    os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
    os.environ['CUDA_VISIBLE_DEVICES'] = '-1'
    # tf.config.experimental.set_memory_growth(tf.config.list_physical_devices("GPU")[0], True)

    batch_size = 16
    sequence_length = 16

    optimizer: tf.keras.optimizers.Optimizer = tf.keras.optimizers.SGD()

    # without vision
    env = make_task("ManipulateBlockDiscreteAsynchronous-v0")
    _, _, joint = build_shadow_v2_brain_base(env, MultiCategoricalPolicyDistribution(env), bs=batch_size, blind=True,
                                          sequence_length=sequence_length, model_type="gru")
    plot_model(joint, to_file=f"{joint.name}.png", expand_nested=True, show_shapes=True)

    joint.summary()

    out = joint({
        "vision": tf.random.normal((batch_size, sequence_length, 7)),
        "proprioception": tf.random.normal((batch_size, sequence_length, 48)),
        "touch": tf.random.normal((batch_size, sequence_length, 92)),
        "asymmetric": tf.random.normal((batch_size, sequence_length, 25)),
        "goal": tf.random.normal((batch_size, sequence_length, 4)),
    })

    # with vision
    env = make_task("VisualManipulateBlockDiscreteAsynchronous-v0")
    _, _, joint = build_shadow_v2_brain_base(env, MultiCategoricalPolicyDistribution(env), bs=batch_size, blind=False,
                                          sequence_length=sequence_length, model_type="gru")
    plot_model(joint, to_file=f"{joint.name}.png", expand_nested=True, show_shapes=True)

    out = joint({
        "vision": tf.random.normal((batch_size, sequence_length, VISION_WH, VISION_WH, 3)),
        "proprioception": tf.random.normal((batch_size, sequence_length, 48)),
        "touch": tf.random.normal((batch_size, sequence_length, 92)),
        "asymmetric": tf.random.normal((batch_size, sequence_length, 25)),
        "goal": tf.random.normal((batch_size, sequence_length, 4)),
    })

    joint.summary()



#!/usr/bin/env python
"""BaseShadowHandEnv Environment Wrappers."""
import abc
from typing import Optional

import gymnasium as gym
import numpy as np
from gymnasium import spaces

from angorapy.common.const import N_SUBSTEPS
from angorapy.tasks.core import AnthropomorphicEnv
from angorapy.tasks.envs.dexterity.consts import DEFAULT_INITIAL_QPOS
from angorapy.tasks.envs.dexterity.consts import FINGERTIP_SITE_NAMES
from angorapy.tasks.envs.dexterity.mujoco_model.robot import ShadowHand
from angorapy.tasks.utils import mj_get_category_names
from angorapy.utilities.core import mpi_print


class BaseShadowHandEnv(AnthropomorphicEnv, abc.ABC):
    """Base class for all shadow hand environments, setting up mostly visual characteristics of the environment."""
    discrete_bin_count = 11

    def __init__(self,
                 initial_qpos=DEFAULT_INITIAL_QPOS,
                 n_substeps=N_SUBSTEPS,
                 delta_t=0.002,
                 relative_control=True,
                 model=None,
                 vision=False,
                 touch=True,
                 render_mode: Optional[str] = None):
        gym.utils.EzPickle.__init__(**locals())

        if model is None:
            model = ShadowHand()

        self._thumb_touch_sensor_id = []
        self.reward_type = "dense"
        self._freeze_wrist = False
        self.color_scheme = "default"
        self.viewpoint = "topdown"
        self.thumb_name = 'robot/rh_S_thtip'
        self.palm_name = 'robot/rh_palm'

        super(BaseShadowHandEnv,
              self).__init__(model=model,
                             frame_skip=n_substeps,
                             initial_qpos=initial_qpos,
                             vision=vision,
                             touch=touch,
                             relative_control=relative_control,
                             render_mode=render_mode,
                             delta_t=delta_t,
                             n_substeps=n_substeps)

    # SPACES
    def _set_observation_space(self, obs):
        # bounds are set to max of dtype to avoid infinity warnings
        self.observation_space = spaces.Dict(dict(
            observation=spaces.Dict(
                {name: spaces.Box(np.finfo(np.float32).min,
                                  np.finfo(np.float32).max,
                                  shape=val.shape)
                 for name, val in obs['observation'].dict().items()}
            ),
        ))

    # TOGGLES
    def toggle_wrist_freezing(self):
        """Toggle flag preventing the wrist from moving."""
        self._freeze_wrist = not self._freeze_wrist
        print("Wrist movements are now frozen.")

    # GETTERS
    def get_finger_position(self,
                            finger_name: str) -> np.ndarray:
        """Get position of the specified finger in space."""
        return self.data.site(finger_name).xpos.flatten()

    def get_fingertip_positions(self):
        """Get positions of all fingertips in euclidean space. Each position is encoded by three floating point numbers,
        as such the output is a 15-D numpy array."""
        goal = [self.data.site(name).xpos.flatten() for name in FINGERTIP_SITE_NAMES]
        return np.array(goal).flatten()

    def get_thumb_position(self) -> np.ndarray:
        """Get the position of the thumb in space."""
        return self.data.site(self.thumb_name).xpos.flatten()

    def get_palm_position(self) -> np.ndarray:
        """Return the robotic hand's palm's center."""
        return self.model.site("robot/palm_center_site").pos

    def is_thumb_tip_touching(self):
        if sum(self.data.sensordata[self._touch_sensor_id]) > 0.0:
            return True

    # CONTROL
    def step(self, action: np.ndarray):
        if self._freeze_wrist:
            action[:2] = 0

        return super(BaseShadowHandEnv, self).step(action)

    # SIMULATION
    def _env_setup(self, initial_state):
        """Initial configuration of the environment. Can be used to configure initial state
        and extract information from the simulation."""
        for k, v in zip(mj_get_category_names(self.model, "sensor")[0], self.model.sensor_adr):
            if b'robot/rh_TS_' in k:
                self._touch_sensor_id.append(v)

                if b'thtip' in k:
                    self._thumb_touch_sensor_id.append(v)

        if initial_state is not None:
            self.set_state(**initial_state)

    # RENDERING
    def change_perspective(self,
                           perspective: str):
        assert perspective in ["topdown", "side", "topdown-far"], "This viewpoint has no settings available."

        self.viewpoint = perspective
        self.viewer_setup()

    def change_color_scheme(self,
                            color_scheme: str):
        assert color_scheme in ["default", "inverted"]

        self.color_scheme = color_scheme
        self.viewer_setup()

    def _get_info(self):
        return super(BaseShadowHandEnv, self)._get_info()

    def viewer_setup(self):
        # hand color
        if self.color_scheme == "default":
            pass
            # self.model.mat_rgba[2] = np.array([30, 30, 30, 255]) / 255  # hand
        elif self.color_scheme == "inverted":
            self.model.mat_rgba[0] = np.array([200, 200, 200, 255]) / 255  # hand
            self.model.mat_rgba[1] = np.array([0, 0, 0, 255]) / 255  # background
        else:
            raise NotImplementedError(f"Unknown Color Scheme {self.color_scheme}.")

        # self.sim.model.mat_rgba[4] = np.array([159, 41, 54, 255]) / 255  # background
        # self.sim.model.geom_rgba[48] = np.array([0.5, 0.5, 0.5, 0])

        if self.viewer is None:
            return

        if self.viewpoint == "topdown":
            # rotate camera to top down view
            self.viewer.cam.distance = 0.5  # zoom in
            self.viewer.cam.azimuth = 0.0  # wrist to the bottom
            self.viewer.cam.elevation = -90.0  # top down view
            self.viewer.cam.lookat[0] = 0.35  # slightly move forward
        elif self.viewpoint == "topdown-far":
            # rotate camera to top down view
            self.viewer.cam.distance = 0.7  # increased values zoom out
            self.viewer.cam.azimuth = -0.0  # wrist to the bottom
            self.viewer.cam.elevation = -90.0  # top down view
            self.viewer.cam.lookat[0] += 0.25  # slightly move forward
        elif self.viewpoint == "side":
            # rotate camera to side view
            self.viewer.cam.distance = 0.35  # zoom in
            self.viewer.cam.azimuth = 25.0  # top down view
            self.viewer.cam.elevation = -45.0  # top down view
            self.viewer.cam.lookat[0] -= 0.04  # slightly move forward
        elif self.viewpoint == "angled":
            self.viewer.cam.distance = 0.5  # zoom in
            self.viewer.cam.azimuth = 35.0  # wrist to the bottom
            self.viewer.cam.elevation = -125.0  # top down view
            self.viewer.cam.lookat[0] += 0.  # slightly move forward
            self.viewer.cam.lookat[1] += 0.0  # slightly move forward
        else:
            raise NotImplementedError(f"Unknown Viewpoint {self.viewpoint}.")

        mpi_print(f"Set up color scheme {self.color_scheme}.")
        mpi_print(f"Set up camera with viewpoint '{self.viewpoint}'.")
        mpi_print("Camera setup finished.")



from angorapy.tasks.envs.dexterity.reward_configs import *


def resolve_config_name(config_name: str):
    """Convert a name of a reward configuration into the corresponding dictionary."""
    parts = config_name.split(".")

    try:
        collection = globals()[parts[0]]
        return collection[parts[1]]
    except KeyError:
        raise KeyError("Reward configuration name could not be resolved.")



from pathlib import Path
from typing import Sequence

from dm_control import mjcf

from angorapy.tasks.world_building.entities import External


class Block(External):

    def __init__(self,
                 position: Sequence[float] = (.0, .0, .0),
                 rotation: Sequence[float] = (.0, .0, .0, .0)):
        super().__init__(mjcf.RootElement(),
                         "block",
                         position=position,
                         rotation=rotation)

    def _parse_entity(self) -> None:
        pass

    def _setup_entity(self) -> None:
        self.mjcf_model.compiler.texturedir = str(Path(__file__).parent / "../textures")

        self.mjcf_model.asset.add(
            "texture",
            name="texgeom",
            type="cube",
            builtin="flat",
            mark="cross",
            width=127,
            height=127,
            rgb1=[0.3, 0.6, 0.5],
        )

        self.mjcf_model.asset.add(
            "texture",
            name="texture:object",
            file="block.png",
            gridsize=[3, 4],
            gridlayout=".U..LFRB.D..")

        self.mjcf_model.asset.add(
            "texture",
            name="texture:hidden",
            file="block_hidden.png",
            gridsize=[3, 4],
            gridlayout=".U..LFRB.D..")

        self.mjcf_model.asset.add(
            "material",
            name="object",
            texture="texgeom",
            texuniform=False
        )

        self.mjcf_model.asset.add(
            "material",
            name="material:object",
            texture="texture:object",
            specular=1,
            shininess=0.3,
            reflectance=0
        )

        self.mjcf_model.asset.add(
            "material",
            name="material:hidden",
            texture="texture:hidden",
            specular=1,
            shininess=0.3,
            reflectance=0,
        )
        self.mjcf_model.asset.add(
            "material",
            name="material:target",
            texture="texture:object",
            specular=1,
            shininess=0.3,
            reflectance=0,
            rgba=[1, 1, 1, 0.5]
        )

        self.mjcf_model.worldbody.add("geom", name="object", type="box", size=[0.025, 0.025, 0.025],
                                      material="material:object",
                                      condim=4,
                                      density=567)
        self.mjcf_model.worldbody.add("geom", name="object_hidden", type="box",
                        size=[0.024, 0.024, 0.024],
                        material="material:hidden",
                        condim=4,
                        contype=0,
                        conaffinity=0,
                        mass=0
                        )
        self.mjcf_model.worldbody.add("site",
                                      name="object:center",
                                      pos=[0, 0, 0],
                                      size=[0.01, 0.01, 0.01],
                                      rgba=[1, 0, 0, 0])

    def add_object_joint(self, attachment_frame: mjcf.Element):
        attachment_frame.set_attributes(
            pos=[0.32, 0.0, 0.1]
        )
        attachment_frame.add("joint", type="free", damping=0.01, name="object:joint")

    @property
    def joints(self) -> Sequence[mjcf.Element]:
        return self._mjcf_root.find_all("joint")

    @property
    def root_body(self) -> mjcf.Element:
        return self.mjcf_model.find("body", "block")



from typing import List

import numpy as np
import tensorflow as tf
import tqdm
from sklearn.linear_model import Ridge

from angorapy.analysis.investigators import base_investigator
from angorapy.common.policies import BasePolicyDistribution
from angorapy.common.senses import stack_sensations
from angorapy.tasks.wrappers import TaskWrapper
from angorapy.utilities.core import flatten, stack_dicts


class Predictability(base_investigator.Investigator):
    """Investigate predictability of information from the activity in different regions of the network."""

    def __init__(self, network: tf.keras.Model, distribution: BasePolicyDistribution):
        super().__init__(network, distribution)
        self._data = None
        self.prepared = False

        self.n_states = np.nan

    def prepare(self, env: TaskWrapper, layers: List[str], n_states=1000, verbose=True):
        """Prepare samples to predict from."""
        self.data, self.train_data, self.val_data = None, None, None

        self.network.reset_states()
        state, info = env.reset(return_info=True)

        state_collection = []
        activation_collection = []
        other_collection = []
        prev_achieved_goals = [info["achieved_goal"]] * 50
        prev_object_quats = [info["achieved_goal"][3:]] * 10
        need_thump_this_episode = False
        for _ in tqdm.tqdm(range(n_states), disable=not verbose):
            state_collection.append(state)
            other = {
                "reward": info.get("original_reward", 0),
            }

            prepared_state = state.with_leading_dims(time=self.is_recurrent).dict()
            activations = self.get_layer_activations(layers, prepared_state)
            activation_collection.append(activations)

            probabilities = flatten(activations["output"])
            action, _ = self.distribution.act(*probabilities)

            observation, reward, terminated, truncated, info = env.step(action)
            done = terminated or truncated

            # if env.is_thumb_tip_touching():
            #     need_thump_this_episode = True

            # other["needed_thumb"] = None
            other["fingertip_positions"] = env.unwrapped.get_fingertip_positions()
            other["translation"] = env.unwrapped._goal_distance(info["achieved_goal"], prev_achieved_goals[-1])
            other["translation_to_10"] = env.unwrapped._goal_distance(info["achieved_goal"], prev_achieved_goals[-10])
            other["translation_to_20"] = env.unwrapped._goal_distance(info["achieved_goal"], prev_achieved_goals[-20])
            other["translation_to_30"] = env.unwrapped._goal_distance(info["achieved_goal"], prev_achieved_goals[-30])
            other["translation_to_40"] = env.unwrapped._goal_distance(info["achieved_goal"], prev_achieved_goals[-40])
            other["translation_to_50"] = env.unwrapped._goal_distance(info["achieved_goal"], prev_achieved_goals[-50])
            other["object_orientation"] = env.unwrapped.data.jnt('block/object:joint/').qpos.copy()[3:]
            # other["relative_angle"] = tfg.quaternion.relative_angle(
            #     tf.cast(env.unwrapped.data.jnt('object:joint').qpos.copy()[3:], dtype=tf.float64),
            #     tf.cast(state["goal"], dtype=tf.float64)
            # )
            # prev_quat = transform.Rotation.from_quat(prev_object_quats[-1])
            # prev_quat_10 = transform.Rotation.from_quat(prev_object_quats[-10])
            # current_quat = transform.Rotation.from_quat(other["object_orientation"])
            # change_in_orientation = (prev_quat_10 * current_quat.inv()).as_quat()

            # other["rotation_matrix"] = quat2mat((prev_quat * current_quat.inv()).as_quat()).flatten()
            # other["rotation_matrix_last_10"] = quat2mat(change_in_orientation).flatten()
            # other["current_rotational_axis"] = quat2axisangle(change_in_orientation)[0]  # todo
            other["goals_achieved_so_far"] = env.consecutive_goals_reached

            other_collection.append(other)

            if done:
                state, info = env.reset(return_info=True)
                prev_achieved_goals = [info["achieved_goal"]] * 50

                self.network.reset_states()

                # for ot in other_collection:
                #     ot["needed_thumb"] = ot["needed_thumb"] if ot["needed_thumb"] is not None else need_thump_this_episode

                # need_thump_this_episode = False

            else:
                state = observation
                prev_achieved_goals.append(info["achieved_goal"])
                # for ot in other_collection:
                #     ot["needed_thumb"] = ot["needed_thumb"] if ot["needed_thumb"] is not None else need_thump_this_episode

                if len(prev_achieved_goals) > 50:
                    prev_achieved_goals.pop(0)

        stacked_other_collection = stack_dicts(other_collection)
        stacked_other_collection["translation"][:-1] = stacked_other_collection["translation"][1:]
        stacked_other_collection["translation_to_10"][:-10] = stacked_other_collection["translation_to_10"][10:]
        stacked_other_collection["translation_to_20"][:-20] = stacked_other_collection["translation_to_20"][20:]
        stacked_other_collection["translation_to_30"][:-30] = stacked_other_collection["translation_to_30"][30:]
        stacked_other_collection["translation_to_40"][:-40] = stacked_other_collection["translation_to_40"][40:]
        stacked_other_collection["translation_to_50"][:-50] = stacked_other_collection["translation_to_50"][50:]

        stacked_other_collection["translation"][-1:] = stacked_other_collection["translation"][-1]
        stacked_other_collection["translation_to_10"][-10:] = stacked_other_collection["translation"][-1]
        stacked_other_collection["translation_to_20"][-20:] = stacked_other_collection["translation"][-1]
        stacked_other_collection["translation_to_30"][-30:] = stacked_other_collection["translation"][-1]
        stacked_other_collection["translation_to_40"][-40:] = stacked_other_collection["translation"][-1]
        stacked_other_collection["translation_to_50"][-50:] = stacked_other_collection["translation"][-1]

        self._data = tf.data.Dataset.from_tensor_slices({
            **stack_dicts(activation_collection),
            **stack_sensations(state_collection).dict(),
            **stacked_other_collection
        }).shuffle(n_states, reshuffle_each_iteration=False)

        self.train_data, self.val_data = self._data.skip(int(0.2 * n_states)), self._data.take(int(0.2 * n_states))
        self.prepared = True

    def fit(self, source_layer: str, target_information: str):
        """Measure the predictability of target_information based on the information in source_layer's activation."""
        assert self.prepared, "Need to prepare before investigating."
        assert source_layer in list(self.train_data.element_spec.keys()) + ["noise"], f"{source_layer} not found"
        assert target_information in list(self.train_data.element_spec.keys())

        predictor = Ridge()

        base_shape = np.squeeze(list(self._data.as_numpy_iterator().__next__().values())[0]).shape
        X = list(self.train_data.map(lambda x: (tf.random.normal(base_shape) if source_layer == "noise" else tf.squeeze(x[source_layer]))).as_numpy_iterator())
        Y = list(self.train_data.map(lambda x: (tf.random.normal(base_shape) if source_layer == "noise" else tf.squeeze(x[target_information]))).as_numpy_iterator())
        X_test = list(self.val_data.map(lambda x: (tf.random.normal(base_shape) if source_layer == "noise" else tf.squeeze(
            x[source_layer]))).as_numpy_iterator())
        Y_test = list(self.val_data.map(lambda x: (tf.random.normal(base_shape) if source_layer == "noise" else tf.squeeze(
            x[target_information]))).as_numpy_iterator())

        predictor.fit(X, Y)
        score = predictor.score(X_test, Y_test)

        # print(f"{target_information} from {source_layer} has an R2 of {predictor.score(X_test, Y_test)}.")

        return score


from angorapy.tasks.world_building.base import PhysicalWorld
from angorapy.tasks.world_building.entities import Stage
from angorapy.tasks.envs.dexterity.mujoco_model.externals import Block
from angorapy.tasks.envs.dexterity.mujoco_model.robot import ShadowHand


class ShadowHandWithCubeWorld(PhysicalWorld):

    def __init__(self):
        self._stage = Stage()
        self._robot = ShadowHand()
        self._cube = Block()

        self._stage.attach(self._robot)
        cube_attachment_frame = self._stage.attach(self._cube)
        self._cube.add_object_joint(cube_attachment_frame)

    @property
    def stage(self) -> Stage:
        return self._stage

    @property
    def robot(self) -> ShadowHand:
        return self._robot

    @property
    def cube(self) -> Block:
        return self._cube


if __name__ == '__main__':
    world = ShadowHandWithCubeWorld()
    print(world.stage.mjcf_model.to_xml_string())


from typing import List, Tuple, Dict

import numpy as np
from numpy.core._multiarray_umath import ndarray as arr

from angorapy.common.senses import Sensation, stack_sensations


class ExperienceBuffer:
    """Buffer for experience gathered by a single worker."""

    def __init__(self, capacity: int, state_dim: Dict[str, Tuple[int]], action_dim: Tuple[int], is_continuous: bool):
        """Initialize the buffer with empty numpy arrays.

        Individual buffers include a leading batch dimension."""

        # data buffers
        self.states = {sense: np.empty((capacity, *shape,), dtype=np.float32) for sense, shape in state_dim.items()}
        self.actions = np.empty((capacity, *action_dim[:-1]), dtype=np.float32 if is_continuous else np.int32)
        self.action_probabilities = np.empty((capacity,), dtype=np.float32)
        self.returns = np.empty((capacity,), dtype=np.float32)
        self.advantages = np.empty((capacity,), dtype=np.float32)
        self.values = np.empty((capacity,), dtype=np.float32)
        self.dones = np.zeros((capacity,), dtype=bool)

        # only here to be compatible with TimeDistributed, no functionality in this implementation
        self.mask = np.ones((capacity,), dtype=bool)

        # data not stored but buffered (for postprocessing etc.)
        self.achieved_goals = []

        # secondary data
        self.episode_lengths = []
        self.episode_rewards = []
        self.episodes_completed = 0
        self.auxiliary_performances = {}

        # indicators
        self.is_continuous = is_continuous
        self.capacity = capacity
        self.filled = 0

        self.seq_length = 1

    def __repr__(self):
        return f"{self.__class__.__name__}[{self.filled}/{self.capacity}]"

    def fill(self, s: List[Sensation], a: arr, ap: arr, adv: arr, ret: arr, v: arr, dones: arr, achieved_goals: arr):
        """Fill (and thereby overwrite) the buffer with a 5-tuple of experience."""
        assert np.all(np.array(list(map(len, [s, a, ap, ret, adv, v]))) == len(s)), f"Inconsistent input sizes: {np.array(list(map(len, [s, a, ap, ret, adv, v])))}"

        self.capacity = len(adv)

        s = stack_sensations(s)
        self.states = {sense: s[sense] for sense in self.states.keys()}
        self.actions = a
        self.action_probabilities = ap
        self.returns = ret
        self.advantages = adv
        self.values = v
        self.dones = dones
        self.achieved_goals = achieved_goals

        # update this to avoid incompatible sizes in the case of changing capacity
        self.mask = np.ones((self.capacity,), dtype=bool)

    def normalize_advantages(self):
        """Normalize the buffered advantages using z-scores. This requires the sequences to be of equal lengths,
        hence if this is not guaranteed during pushing to the buffer, pad_buffer has to be called first."""
        rank = len(self.advantages.shape)
        assert rank in [1, 2], f"Illegal rank of advantage tensor, should be 1 or 2 but is {rank}."

        mean = np.mean(self.advantages)
        std = np.maximum(np.std(self.advantages), 1e-6)
        self.advantages = (self.advantages - mean) / std


class TimeSequenceExperienceBuffer(ExperienceBuffer):
    """Experience Buffer for TimeSequence Data

    Attributes:
        states:         dims (B,
    """

    def __init__(self, capacity: int, state_dim: Dict[str, Tuple[int]], action_dim: Tuple[int], is_continuous: bool,
                 seq_length: int):

        super().__init__(capacity, state_dim, action_dim, is_continuous)
        self.seq_length = seq_length

        self.states = {sense: np.zeros((1, capacity, self.seq_length, *shape,), dtype=np.float32) for sense, shape in
                       state_dim.items()}
        self.actions = np.zeros((1, capacity, self.seq_length, *action_dim[:-1],), dtype=np.float32 if is_continuous else np.int32)
        self.action_probabilities = np.zeros((1, capacity, self.seq_length,), dtype=np.float32)
        self.returns = np.zeros((1, capacity, self.seq_length,), dtype=np.float32)
        self.advantages = np.zeros((1, capacity, self.seq_length,), dtype=np.float32)
        self.values = np.zeros((1, capacity, self.seq_length,), dtype=np.float32)

        self.true_number_of_transitions = 0
        self.number_of_subsequences_pushed = 0

        self.dones = np.zeros((1, capacity, self.seq_length,), dtype=bool)
        self.mask = np.zeros(self.advantages.shape, dtype=bool)

        self.last_advantage_stop = 0

    def __repr__(self):
        return f"{self.__class__.__name__}[{self.filled}/{self.capacity} - {self.seq_length}]"

    def push_seq_to_buffer(self, states: List[Sensation], actions: List[arr], action_probabilities: List[arr],
                           values: List[arr], episode_ended: bool = False):
        """Push a sequence to the buffer, constructed from given lists of values."""
        # assert np.all(np.array([len(states), len(actions), len(action_probabilities), len(values)]) == len(states)), \
        #     "Inconsistent input sizes."

        seq_length = len(actions)

        if seq_length < self.seq_length and not episode_ended:
            raise Exception("An incomplete subsequence has been pushed to the buffer without flagging the episode"
                            "as finished during the subsequence.")

        states = stack_sensations(states)
        for sense, sense_buffer in self.states.items():
            sense_buffer[0, self.number_of_subsequences_pushed, :seq_length, ...] = states[sense]

        self.actions[0, self.number_of_subsequences_pushed, :seq_length, ...] = np.stack(actions)
        self.action_probabilities[0, self.number_of_subsequences_pushed, :seq_length] = action_probabilities
        self.values[0, self.number_of_subsequences_pushed, :seq_length] = values

        if episode_ended:
            self.dones[0, self.number_of_subsequences_pushed, seq_length - 1] = 1

        self.number_of_subsequences_pushed += 1
        self.filled += self.actions.shape[1]
        self.true_number_of_transitions += seq_length

    def push_adv_ret_to_buffer(self, advantages: arr, returns: arr):
        """Push advantages and returns of a whole episode."""
        overhang = len(advantages) % self.seq_length

        # split advantages if necessary
        if overhang == 0:
            advantage_chunks = np.split(advantages, indices_or_sections=len(advantages) // self.seq_length)
            return_chunks = np.split(returns, indices_or_sections=len(advantages) // self.seq_length)
        elif len(advantages) > self.seq_length:
            advantage_chunks = (np.split(advantages[:-overhang], indices_or_sections=len(advantages) // self.seq_length)
                                + [advantages[-overhang:]])
            return_chunks = (np.split(returns[:-overhang], indices_or_sections=len(advantages) // self.seq_length)
                             + [returns[-overhang:]])
        else:
            advantage_chunks = [advantages]
            return_chunks = [returns]

        # fill in the subsequences one by on
        for adv_sub_seq, ret_sub_seq in zip(advantage_chunks, return_chunks):
            seq_length = len(adv_sub_seq)
            self.advantages[0, self.last_advantage_stop, :seq_length] = adv_sub_seq
            self.mask[0, self.last_advantage_stop, :seq_length] = np.ones(adv_sub_seq.shape)
            self.returns[0, self.last_advantage_stop, :seq_length] = ret_sub_seq

            self.last_advantage_stop += 1

    def normalize_advantages(self):
        """Normalize the buffered advantages using z-scores. This requires the sequences to be of equal lengths."""
        rank = len(self.advantages.shape)
        assert rank in [1, 2, 3], f"Illegal rank of advantage tensor, should be 1 or 2 but is {rank}."

        # TODO could better keep track of already masked array in self.advantages?
        masked_advantages = np.ma.masked_array(self.advantages, 1 - self.mask)
        mean = masked_advantages.mean()
        std = np.maximum(masked_advantages.std(), 1e-6)
        self.advantages = (self.advantages - mean) / std



"""Helper functions."""
import os
import random
import sys
from typing import Dict
from typing import List
from typing import Tuple
from typing import Union

import gymnasium as gym
import numpy
import numpy as np
import tensorflow as tf
from gymnasium import spaces
from gymnasium.spaces import Box
from gymnasium.spaces import Discrete
from gymnasium.spaces import MultiDiscrete

try:
    from mpi4py import MPI
except:
    MPI = None

from angorapy.utilities.error import UninterpretableObservationSpace


def mpi_flat_print(string: str):
    """A bit of a workaround to no new line printing to have it work in PyCharm."""
    if MPI is None or MPI.COMM_WORLD.Get_rank() == 0:
        print(f"\r{string}", end="")


def mpi_print(string: str):
    """A bit of a workaround to no new line printing to have it work in PyCharm."""
    if MPI is None or MPI.COMM_WORLD.Get_rank() == 0:
        print(string)


def set_all_seeds(seed):
    """Set all random seeds (tf, np, random) to given value."""
    tf.random.set_seed(seed)
    np.random.seed(seed)
    random.seed(seed)


def env_extract_dims(env: gym.Env) -> Tuple[Dict[str, Tuple], Tuple[int]]:
    """Returns state and action space dimensionality for given environment."""
    obs_dim: dict

    # extract observation space dimensionalities from environment
    if isinstance(env.observation_space, spaces.Dict):
        # dict observation with observation field, for all GoalEnvs
        if isinstance(env.observation_space["observation"], spaces.Box):
            obs_dim = {"proprioception": env.observation_space["observation"].shape,
                       "goal": env.observation_space["desired_goal"].shape}
        elif isinstance(env.observation_space["observation"], spaces.Dict):
            # made for sensation
            obs_dim = {name: field.shape for name, field in env.observation_space["observation"].spaces.items()}
        else:
            raise UninterpretableObservationSpace(
                f"Cannot extract the dimensionality from a spaces.Dict observation space "
                f"where the observation is of type "
                f"{type(env.observation_space['observation']).__name__}")
    elif isinstance(env.observation_space, gym.spaces.Box):  # standard observation in box form
        obs_dim = {"proprioception": env.observation_space.shape}
    elif isinstance(env.observation_space, gym.spaces.tuple.Tuple):
        obs_dim = {"proprioception": (len(env.observation_space),)}
    else:
        raise UninterpretableObservationSpace(
            f"Cannot interpret observation space of type {type(env.observation_space)}.")

    # action space
    if isinstance(env.action_space, Discrete):
        act_dim = (int(env.action_space.n),)
    elif isinstance(env.action_space, MultiDiscrete):
        assert np.alltrue(env.action_space.nvec == env.action_space.nvec[0]), "Can only handle multi-discrete action" \
                                                                              "spaces where all actions have the same " \
                                                                              "number of categories."
        act_dim = (int(env.action_space.shape[0]), int(env.action_space.nvec[0].item()))
    elif isinstance(env.action_space, Box):
        act_dim = (int(env.action_space.shape[0]), 1)
    else:
        raise NotImplementedError(f"Environment has unknown Action Space Typ: {env.action_space}")

    return obs_dim, act_dim


@tf.function
def normalize(x, is_img=False) -> numpy.ndarray:
    """Normalize a numpy array to have all values in range (0, 1)."""
    x = tf.convert_to_tensor(x).numpy()
    return x / 255 if is_img else (x - x.min()) / (x.max() - x.min())


def flatten(some_list):
    """Flatten a python list."""
    return [some_list] if not isinstance(some_list, list) else [x for X in some_list for x in flatten(X)]


def is_array_collection(a: numpy.ndarray) -> bool:
    """Check if an array is an array of objects (e.g. goal arrays) or an actual array of direct data."""
    return a.dtype == "O"


def add_state_dims(state: "Sensation", dims: int = 1, axis: int = 0) -> 'Sensation':
    """Expand state (array or lost of arrays) to have a batch and/or time dimension."""
    if dims < 1:
        return state

    new_state = state.with_leading_dims(time=dims == 2)

    return new_state


def merge_into_batch(list_of_states: List[Union[numpy.ndarray, Tuple]]):
    """Merge a list of states into one huge batch of states. Handles both single and multi input states.

    Assumes NO batch dimension!
    """
    if isinstance(list_of_states[0], numpy.ndarray):
        return numpy.concatenate(add_state_dims(list_of_states))
    else:
        return tuple(numpy.concatenate(list(map(lambda x: add_state_dims(x[i]), list_of_states)), axis=0)
                     for i in range(len(list_of_states[0])))


def insert_unknown_shape_dimensions(shape, none_replacer: int = 1):
    """Replace Nones in a shape tuple with 1 or a given goal value."""
    return tuple(map(lambda s: none_replacer if s is None else s, shape))


def detect_finished_episodes(dones: tf.Tensor):
    """Detect which samples in the batch connect to a episode that finished during the subsequence, based on the dones
     and return a 1D boolean tensor.

    Input Shape:
        dones: (B, S)
    """
    finished = tf.math.reduce_any(dones, axis=-1)
    return finished


def find_divisors(number: int):
    divisors = []
    for i in range(1, number // 2 + 1):
        if number % i == 0:
            divisors.append(i)
            divisors.append(number // i)

    return list(sorted(set(divisors))) + [number]


def find_optimal_tile_shape(floor_shape: Tuple[int, int], tile_size: int, width_first=False) -> Tuple[int, int]:
    """For a given shape of a matrix (floor), find the shape of a tiles that fit the floor and contain
    exactly tile_size elements."""
    height_divisors = find_divisors(floor_shape[0])
    width_divisors = find_divisors(floor_shape[1])

    if not width_first:
        height_divisors = list(reversed(height_divisors))
    else:
        width_divisors = list(reversed(width_divisors))

    for hd in height_divisors:
        for wd in width_divisors:
            if hd * wd == tile_size:
                return hd, wd

    raise ValueError(f"No tiling of size {tile_size} possible for a floor of shape {floor_shape}.")


def stack_dicts(dicts: List[Dict[str, tf.Tensor]]):
    """Stack matching elements of a dict over a prepended domain."""
    return {
        key: np.stack([s[key] for s in dicts], axis=0) for key in dicts[0].keys()
    }


class HiddenPrints:
    """Context that hides print calls."""

    def __enter__(self):
        self._original_stdout = sys.stdout
        sys.stdout = open(os.devnull, 'w')

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stdout.close()
        sys.stdout = self._original_stdout


def keras_model_memory_usage_in_bytes(model, *, batch_size: int):
    """
    Return the estimated memory usage of a given Keras model in bytes.
    This includes the model weights and layers, but excludes the dataset.
    The model shapes are multipled by the batch size, but the weights are not.
    Args:
        model: A Keras model.
        batch_size: The batch size you intend to run the model with. If you
            have already specified the batch size in the model itself, then
            pass `1` as the argument here.
    Returns:
        An estimate of the Keras model's memory usage in bytes.
    """
    default_dtype = tf.keras.backend.floatx()
    shapes_mem_count = 0
    internal_model_mem_count = 0
    for layer in model.layers:
        if isinstance(layer, tf.keras.Model):
            internal_model_mem_count += keras_model_memory_usage_in_bytes(
                layer, batch_size=batch_size
            )
        single_layer_mem = tf.as_dtype(layer.dtype or default_dtype).size
        out_shape = layer.output_shape
        if isinstance(out_shape, list):
            out_shape = out_shape[0]
        for s in out_shape:
            if s is None:
                continue
            single_layer_mem *= s
        shapes_mem_count += single_layer_mem

    trainable_count = sum(
        [tf.keras.backend.count_params(p) for p in model.trainable_weights]
    )
    non_trainable_count = sum(
        [tf.keras.backend.count_params(p) for p in model.non_trainable_weights]
    )

    total_memory = (
            batch_size * shapes_mem_count
            + internal_model_mem_count
            + trainable_count
            + non_trainable_count
    )
    return total_memory
